import lxml
from lxml import etree
from PIL import Image, ImageTk
from .widget import Widget, register_widget

class WT_FINGER_CATCHER(Widget):
    """
        "finger catcher" for digging events
    """
    def __init__(this, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

register_widget('WT_FINGER_CATCHER', WT_FINGER_CATCHER,)



# Thanks to @campbellsonic for the wrapRawData script. It was originally written in C#, I just rewrote it in python. I couldn't have done it without them.

# usage:
# waltex.py --help

# import numpy
from PIL import Image, ImageFile
import math
import struct
import io
import sys
# import json

try:
    import filetype

    class _WaltexFile(filetype.Type):
        MIME = 'image/waltex'
        EXTENSION = 'waltex'

        def __init__(self):
            """Waltex file extension class for the `filetype` module
            """
            super(_WaltexFile, self).__init__(
                mime=_WaltexFile.MIME,
                extension=_WaltexFile.EXTENSION,
            )

        def match(self, buf):
            return (len(buf) > 3 and
                    buf[0] == 0x57 and
                    buf[1] == 0x41 and
                    buf[2] == 0x4C and
                    buf[3] == 0x54)

    filetype.add_type(_WaltexFile())
# filetype.guess()
except:
    # optional filetype addition
    pass

# add waltex image to PIL
# Thanks to Mark Setchell for most of this code https://stackoverflow.com/a/75511423/17129659
def _accept(prefix):
    return prefix[:4] == b"WALT"

class _WaltexImageFile(ImageFile.ImageFile):

    HEADER_LENGTH = 16
    HEADER_MAGIC = '4sBBHH6s'
    format = "WALTEX"
    format_description = "Waltex walaber texture image"

    def _open(self):
        header = self.fp.read(self.HEADER_LENGTH)
        magic, vers, fmt, w, h, _ = struct.unpack(self.HEADER_MAGIC, header)

        # size in pixels (width, height)
        self._size = w, h

        # mode setting
        self.mode = 'RGBA'

        # Decoder
        if fmt == 0:
            # RGBA8888
            # Just use built-in raw decoder
            self.tile = [("raw", (0, 0) + self.size, self.HEADER_LENGTH, (self.mode, 0, 1))]
        elif fmt == 3:
            # RGBA4444
            # Use raw decoder with custom RGBA;4B unpacker
            self.tile = [("raw", (0, 0) + self.size, self.HEADER_LENGTH, ('RGBA;4B', 0, 1))]
        elif fmt in [1,2]:
            raise NotImplementedError(f'Format {fmt} is not supported yet.\nPlease go to https://github.com/wmw-modding/wmwpy/issues and report the issue. Please include the file too.')

# regester file type in PIL
Image.register_open(_WaltexImageFile.format, _WaltexImageFile, _accept)

Image.register_extensions(
    _WaltexImageFile.format,
    [
        ".waltex"
    ],
)

# main class
 
class Waltex():
    format = 'waltex'
    format_description = "Walaber image file"

    def __init__(
        this,
        file : str | bytes,
        byte_order : str = 'little'
    ) -> None:
        """Waltex image

        Args:
            file (str | bytes): Waltex file. Can be path to file, contents of file as bytes, or file-like object.
            byte_order (str, optional): 'little' or 'big'. Defaults to 'little'.

        Raises:
            TypeError: file has to be a 'str', 'bytes' or file-like object.
        """
        
        this._colorspecs = [
            {
                'order': 'rgba',
                'bpp': [8,8,8,8],
                'spec': 'rgba8888',
            },
            {
                'order': 'rgb',
                'bpp': [5,6,5],
                'spec': 'rgb565',
            },
            {
                'order': 'rgba',
                'bpp': [5,5,5,1],
                'spec': 'rgba5551',
            },
            {
                'order': 'rgba',
                'bpp': [4,4,4,4],
                'spec': 'rgba4444',
            },
        ]
        
        if isinstance(file, (str)):
            with open(file, 'rb') as f:
                rawdata = f.read()
                
        elif isinstance(file, bytes):
            rawdata = file
        
        elif hasattr(file, 'read'):
            rawdata = file.read()
            if isinstance(rawdata, str):
                rawdata = rawdata.encode()
        else:
            raise TypeError(f"file has to be a 'str', 'bytes' or file-like object.")
        
        this._byte_order = byte_order
        
        this.file = file
        this.rawdata = io.BytesIO(rawdata)
        
        this.read(byte_order=this._byte_order)
    
    def read(this, byte_order : str = None) -> Image.Image:
        """Read the waltex image.

        Args:
            byte_order (str, optional): The byte order. Can be 'little' or 'big'. Defaults to None.

        Returns:
            PIL.Image.Image: PIL Image object
        """
        if byte_order:
            this._byte_order = byte_order
            
        header = this.rawdata.getvalue()[0:16]
            
        this.format = int(header[5])
        this.colorspec = this._colorspecs[this.format]
        this.version = int(header[4])
        size = (int.from_bytes(header[6:8], byteorder='little'), int.from_bytes(header[8:10], byteorder='little'))
        this.image = Image.new('RGBA', size)
        
        try:
            this.image = Image.open(this.rawdata)
            if byte_order == 'little':
                A, B, G, R = this.image.split()
                this.image = Image.merge('RGBA', (R,G,B,A))
        except:
            this.image = WaltexImage(this.rawdata, byte_order = byte_order)
        
        this.image.format = _WaltexImageFile.format
        this.image.format_description = _WaltexImageFile.format_description
        return this.image
    
    @property
    def size(this):
        """Return size of image

        Returns:
            tuple[int,int]: tuple(width, height)
        """
        return this.image.size

# legacy functions for reading waltex image
def WaltexImage(path : str | bytes | io.BytesIO, premultiplyAlpha : bool = False, dePremultiplyAlpha : bool = False, endian : str = 'little') -> Image.Image:
    """
    ### Depracted
    Use `Image.load()` instead.
    
    I am only keeping this just in case you need to load RGB565 (01) or RGBA5551 (02) waltex format.
    
    ---
    
    Get image from `waltex` file

    Data on image can be found in coorisponding `imagelist` or in `Data/TextureSettings.xml`.
    
    Args:
        path (str): Path to `waltex` image
        premultiplyAlpha (bool, optional): Defaults to False.
        dePremultiplyAlpha (bool, optional): Defaults to False.
        endian (str, optional): Endian mode. Set to 'big' to use big endian. Defaults to 'little'.

    Returns:
        PIL.Image.Image: Pillow image.
    ---
    Thanks to @campbellsonic for the `WrapRawData()` function.
    """
    
    colorOrder = ''
    bytesPerPixel = 0
    bpprgba = []
    
    # print(colorspace, bytesPerPixel, colorOrder, bpprgba)
    
    if isinstance(path, str):
        with open(path, 'rb') as file:
            rawdata = file.read()
    elif isinstance(path, bytes):
        rawdata = path
    elif hasattr(path, 'read'):
        rawdata = path.read()
    else:
        raise TypeError(f"file has to be a 'str', 'bytes' or file-like object.")
    
        
    # print(filetype.guess(rawdata))
        
    if filetype.guess(rawdata).MIME != 'image/waltex':
        raise TypeError('File is not a waltex')
    version = int(rawdata[4])
    format = int(rawdata[5])
    w = int.from_bytes(rawdata[6:8], byteorder='little')
    h = int.from_bytes(rawdata[8:10], byteorder='little')
    pading = rawdata[10:16]
    
    colorspecs = [
        {
            'order': 'rgba',
            'bpp': [8,8,8,8],
            'spec': 'rgba8888'
        },
        {
            'order': 'rgb',
            'bpp': [5,6,5,0],
            'spec': 'rgb565'
        },
        {
            'order': 'rgba',
            'bpp': [5,5,5,1],
            'spec': 'rgba5551'
        },
        {
            'order': 'rgba',
            'bpp': [4,4,4,4],
            'spec': 'rgba4444'
        },
    ]
    
    spec = colorspecs[format]
    colorOrder = spec['order']
    bpprgba = spec['bpp']
    bytesPerPixel = round(sum(bpprgba) / 8)
    
    image = WrapRawData(rawdata, w, h, bytesPerPixel, bpprgba[0], bpprgba[1], bpprgba[2], bpprgba[3], colorOrder, premultiplyAlpha, dePremultiplyAlpha, 16, endian)
    
    image.version = version
    image.spec = spec
    
    return image

def WrapRawData(rawData : bytes, width : int, height : int, bytesPerPixel : int, redBits : int, greenBits : int, blueBits : int, alphaBits : int, colorOrder : str, premultiplyAlpha : bool = False, dePremultiplyAlpha : bool = False, offset : int = 0, endian = 'little'):
    _8BIT_MASK = 256.0
    OUTBITDEPTH = 8
    DEBUG_MODE = False
    
    colorOrder = colorOrder.lower()
    if endian == 'little':
        colorOrder = colorOrder[::-1]
    
    # width and height are switched due to how PIL creates an image from array
    # image = [[(0, 0, 0, 0)] * height] * width
    image = Image.new('RGBA', (width, height), (0, 0, 0, 0))
    x = 0
    y = 0
    
    redMask = GenerateBinaryMask(redBits)
    greenMask = GenerateBinaryMask(greenBits)
    blueMask = GenerateBinaryMask(blueBits)
    alphaMask = GenerateBinaryMask(alphaBits)
    
    redMax = math.pow(2, redBits)
    greenMax = math.pow(2, greenBits)
    blueMax = math.pow(2, blueBits)
    alphaMax = math.pow(2, alphaBits)
    
    # determine number of loops needed to get every pixel
    numLoops = int(len(rawData) / bytesPerPixel)
    
    # loop for each set of bytes (one pixel)
    for i in range(numLoops):
        pixel = 0
        
        # read all bytes for this pixel
        for j in range(bytesPerPixel):
            nextByte = rawData[i * bytesPerPixel + j + offset]
            
            # print(f'Read byte: {hex(nextByte)}')
            
            # move the byte up
            # if (reverseBytes)
            nextByte = nextByte << (8 * j)
            # else
            # pixel = pixel << 8
            
            # append the next one
            pixel += nextByte
            # print(f'Pixel is now: {hex(pixel)}')
            
        # print(f'Pixel: {pixel}')
        
        # get RGBA values from the pixel
        r = g = b = a = 0
        
        # loop for each channel
        for color in colorOrder:
            
            if color == 'r':
                r = pixel & redMask
                pixel = pixel >> redBits
                
            elif color == 'g':
                g = pixel & greenMask
                pixel = pixel >> greenBits
            
            elif color == 'b':
                b = pixel & blueMask
                pixel = pixel >> blueBits
                
            else:
                a = pixel & alphaMask
                pixel = pixel >> alphaBits
                
        # print(f'Before scale:\nR: {r} G: {g} B: {b} A: {a}')
        
        # scale colors to 8-bit depth (not sure which method is better)
        
        # via floating point division
        if (redMax > 1):
            r = round(r * ((_8BIT_MASK - 1) / (redMax - 1)))
        if (greenMax > 1):
            g = round(g * ((_8BIT_MASK - 1) / (greenMax - 1)))
        if (blueMax > 1):
            b = round(b * ((_8BIT_MASK - 1) / (blueMax - 1)))
        if (alphaMax > 1):
            a = round(a * ((_8BIT_MASK - 1) / (alphaMax - 1)))
        
        # via bit shifting
        # rShift = OUTBITDEPTH - redBits
        # gShift = OUTBITDEPTH - greenBits
        # bShift = OUTBITDEPTH - blueBits
        # aShift = OUTBITDEPTH - alphaBits
        # r = (r << rShift) + (r >> (redBits - rShift))
        # g = (g << gShift) + (r >> (greenBits - gShift))
        # b = (b << bShift) + (r >> (blueBits - bShift))
        # a = (a << aShift) + (a >> (alphaBits - aShift))
        
        # print(f'After scale:\nR: {r} G: {g} B: {b} A: {a}')
        
        # if there are no bits allotted for an alpha channel, make pixel opaque rather than invisible
        if alphaBits == 0:
            a = 255
            
        # a = 255
            
        if dePremultiplyAlpha:
            r = round(r * a / 255.0)
            g = round(g * a / 255.0)
            b = round(b * a / 255.0)
            
        if premultiplyAlpha:
            if (a != 0):
                r = round(r * 255.0 / a)
                g = round(g * 255.0 / a)
                b = round(b * 255.0 / a)
            
        # set the pixel
        rgba = (int(r), int(g), int(b), int(a))
        # print(rgba)
        # image[x][y] = rgba
        image.putpixel((x,y), rgba)
        
        # break after first pixel if in debug mode
        
        
        # iterate coordinates
        x += 1
        if (x == width):
            x = 0
            y += 1
            # if (y > (height - 300) or y % 100 == 0):
            #     print(f'Line {y} of {height} done')
            #     if (DEBUG_MODE):
            #         break
                
        # if there's extra data (like the door overlays in the lawns), stop once the height is reached
        if y == height:
            break
        
    
    return image

def GenerateBinaryMask(numOnes):
    binaryMask = 0
    for i in range(numOnes):
        binaryMask *= 2
        binaryMask += 1
        
    return binaryMask

# test code

if __name__ == "__main__":
    help = 'usage:\n    waltex.py <input> [arguments]\n\narguments:\n    --help -h     displays this help message\n\n    --output -o   output of image\n                    --output <filename>\n\n    --show -s     shows image with default image viewer\n'
    helpargs = ['--help', '-h']
    
    if sys.argv[1] in helpargs:
        print(help)
    else:
        path = sys.argv[1]
        # with open(path, 'rb') as file:
        #     rawData = file.read()

        # image = WaltexImage(path)
        # image.show()

        # new fast method
        image = Waltex(path)
        args = sys.argv[2::]

        if len(args) > 0:
            for a in range(len(args)):
                arg = args[a]
                if arg in ['--output', '-o']:
                    image.image.save(args[a + 1])
                elif arg in ['--show', '-s']:
                    image.image.show()
        else:
            image.image.show()



import typing

from .game import Game

# Where's My Water?
class WMW(Game):
    _DB = '/Data/water.db'
    _BASEASSETS = '/'
    _PROFILE = None
    
    game = 'WMW'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Data/water.db',
        profile: str = None,
        baseassets: str = '/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        
        
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

class WMWF(WMW):
    _DB = '/Data/water-Lite.db'
    
    game = 'WMWF'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Data/water-Lite.db',
        profile: str = None,
        baseassets: str = '/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

# Where's My Perry?
class WMP(Game):
    _DB = '/Perry/Data/perry.db'
    _BASEASSETS = '/Perry/'
    
    game = 'WMP'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Perry/Data/water.db',
        profile: str = None,
        baseassets: str = '/Perry/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

class WMPF(WMP):
    _DB = '/Perry/Data/perry-Lite.db'
    
    game = 'WMPF'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Perry/Data/water-Lite.db',
        profile: str = None,
        baseassets: str = '/Perry/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

# Where's My Mickey?
class WMM(Game):
    _DB = '/Mickey/Data/perry.db'
    _BASEASSETS = '/Mickey/'
    
    game = 'WMM'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Mickey/Data/perry.db',
        profile: str = None,
        baseassets: str = '/Mickey/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

class WMMF(WMM):
    _DB = '/Mickey/Data/perry-Lite.db'
    
    game = 'WMMF'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Mickey/Data/perry-Lite.db',
        profile: str = None,
        baseassets: str = '/Mickey/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

# Where's My XiYangYang?
class WMXYY(Game):
    _DB = '/Perry/Data/perry.db'
    _BASEASSETS = '/Perry/'
    
    game = 'WMXYY'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Perry/Data/water.db',
        profile: str = None,
        baseassets: str = '/Perry/',
        hook: typing.Callable[[int, str, int], typing.Any] = None,
        databasekey = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

class WMWFXYY(Game):
    _DB = '/Perry/Data/perry.db'
    _BASEASSETS = '/Perry/'
    
    game = 'WMWFXYY'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Perry/Data/water.db',
        profile: str = None,
        baseassets: str = '/Perry/',
        hook: typing.Callable[[int, str, int], typing.Any] = None,
        databasekey = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

# Where's My Water 2?
class WMW2(Game):
    _DB = '/Water/Data/perry.db'
    _BASEASSETS = '/Water/'
    _PROFILE = '/Water/Data/factory_profile.json'
    
    game = 'WMW2'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Water/Data/perry.db',
        profile: str = '/Water/Data/factory_profile.json',
        baseassets: str = '/Water/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

# Where's My Holiday?
class WMH(Game):
    _DB = '/Data/water-Lite.db'
    _BASEASSETS = '/'
    
    game = 'WMH'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/assets',
        db: str = '/Data/water-Lite.db',
        profile: str = None,
        baseassets: str = '/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)
    
    def mode(this, mode : typing.Literal['wmw', 'wmp'] = 'wmw'):
        """Switch game mode from 'wmw' to 'wmp' and vice verca.

        Args:
            mode (str, optional): Mode. Can be 'wmw' or 'wmp'. Defaults to 'wmw'.
        """
        
        mode = mode.lower()
        if mode == 'wmw':
            this.db = '/Data/water-Lite.db'
            this.baseassets = '/'
            
            this._DB = '/Data/water-Lite.db'
            this._BASEASSETS = '/'
        elif mode == 'wmp':
            this.db = '/Perry/Data/perry-Lite.db'
            this.baseassets = '/Perry'
            
            this._DB = '/Perry/Data/perry-Lite.db'
            this._BASEASSETS = '/Perry'

        this.filesystem.baseassets = this.baseassets

class WMS(WMPF):
    game = 'WMS'
    
    def __init__(
        this,
        gamepath: str,
        assets: str = '/Content',
        db: str = '/Perry/Data/water-Lite.db',
        profile: str = None,
        baseassets: str = '/Perry/',
        hook: typing.Callable[[int, str, int], typing.Any] = None
    ) -> None:
        super().__init__(gamepath, assets, db, profile, baseassets, hook)

GAMES : dict[str, Game] = {}

def register_game(name : str, class_ : Game):
    """Register a game template

    Args:
        name (str): Name of the game. E.G. 'WMW'
        class_ (Game): A class that has been inherited by the `Game` object.

    Raises:
        TypeError: class has to be inherited by Game
        TypeError: name must be a string
        NameError: game already exists
    """
    
    if not isinstance(class_, type):
        class_ = class_.__class__
    
    if not isinstance(class_(None), Game):
        raise TypeError('class has to be inherited by Game')
    if not isinstance(name, str):
        raise TypeError('name must be a string')
    try:
        if GAMES[name]:
            raise NameError(f'game "{name}" already exists')
    except:
        pass
    GAMES[name] = class_

register_game('WMW', WMW)
register_game('WMWF', WMWF)
register_game('WMP', WMP)
register_game('WMPF', WMPF)
register_game('WMM', WMM)
register_game('WMMXL', WMM)
register_game('WMMF', WMMF)
register_game('WMXYY', WMXYY)
register_game('WMWFXYY', WMWFXYY)
register_game('WMW2',WMW2)
register_game('WMH',WMH)
register_game('WMV',WMH)
register_game('WMS',WMS)



import io
from lxml import etree
from PIL import Image, ImageTk
import numpy
import typing
import copy

from ..Utils.filesystem import *
from ..Utils.logging_utils import log_exception
from .object import Object
from ..gameobject import GameObject

class Level(GameObject):
    XML_TEMPLATE = b"""<?xml version="1.0"?>
    <Objects>
    </Objects>
    """
    
    IMAGE_TEMPLATE = Image.new('P', (90,127), 'white').quantize(colors=256)
    IMAGE_FORMAT = 'PNG'
    
    def __init__(
        this,
        xml : str | bytes | File = None,
        image : str | bytes | File = None,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None,
        assets : str = '/assets',
        baseassets : str = '/',
        load_callback : typing.Callable[[int, str, int], typing.Any] = None,
        ignore_errors : bool = False,
    ) -> None:
        """Load level

        Args:
            xml (str | bytes | File): XML file for level.
            image (str | bytes | File): Image file for level.
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
            assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
        """
        
        this.gamepath = gamepath
        this.assets = assets
        this.filename = ''
        if this.assets == None:
            this.assets = '/assets'
        
        super().__init__(filesystem, gamepath, assets, baseassets)
        
        this.xml_file = super().get_file(xml, template = this.XML_TEMPLATE)
        
        this.xml = etree.parse(this.xml_file).getroot()
        
        this.image_file = super().get_file(image)
        if this.image_file == None:
            this.image = this.IMAGE_TEMPLATE.copy()
        else:
            this.image = Image.open(this.image_file).quantize(colors=256)
        
        this.objects : list[Object] = []
        this.properties : dict[str,str] = {}
        this.challenges : list[Level.Challenge] = []
        this.room = (0,0)
        
        this.read(load_callback = load_callback, ignore_errors = ignore_errors)
        
        this.scale = 5
    
    @property
    def size(this) -> tuple[int,int]:
        """Level image size

        Returns:
            tuple[int,int]: (width,height)
        """
        return this._image.size
    
    @property
    def image(this) -> Image.Image:
        """Scaled up Level image

        Returns:
            PIL.Image.Image: PIL Image
        """
        image = this._image.copy()
        
        size = numpy.array(image.size)
        size = size * this.scale
        
        image = image.resize(size, resample = Image.NEAREST)
        
        return image
        
    @image.setter
    def image(this, value : Image.Image):
        this._image = value
    
    @property
    def PhotoImage(this) -> ImageTk.PhotoImage:
        """Tkinter PhotoImage of the Level image

        Returns:
            ImageTk.PhotoImage: Tkinter PhotoImage
        """
        this._PhotoImage = ImageTk.PhotoImage(this.image)
        return this._PhotoImage

    @property
    def scale(this) -> int:
        """Level size scale
        """
        return this._scale
    @scale.setter
    def scale(this, value : int):
        this._scale = value
        
        for obj in this.objects:
            obj.scale = this._scale
    
    def read(
        this,
        load_callback : typing.Callable[[int, str, int], typing.Any] = None,
        ignore_errors : bool = False,
    ):
        """Read level XML
        """
        this.objects : list[Object] = []
        this.properties = {}
        
        def run_callback(index, name, max):
            try:
                if callable(load_callback):
                    load_callback(index, name, max)
            except:
                log_exception()
        
        id = 0
        
        max = len(this.xml)
        
        index = 0
        
        for element in this.xml:
            element : etree.ElementBase
            
            try:
                
                # comment safe-guard
                if element is etree.Comment:
                    run_callback(index, 'Comment', max)
                    continue
                
                if element.tag == 'Object':
                    properties = {}
                    pos = (0,0)
                    name = element.get('name')
                    
                    run_callback(index, f'Object: {name}', max)
                    
                    for el in element:
                        if not el is etree.Comment:
                            if el.tag == 'AbsoluteLocation':
                                pos = el.get('value')
                            
                            elif el.tag == 'Properties':
                                for property in el:
                                    if not property is etree.Comment and property.tag == 'Property':
                                        properties[property.get('name')] = property.get('value')
                    
                    obj = Object(
                        this.filesystem.get(properties['Filename']), # get file because `Object` does not take filepath
                        filesystem = this.filesystem,
                        properties = properties,
                        pos = pos,
                        name = name,
                    )
                    
                    obj.id = id
                    
                    this.objects.append(obj)
                    
                    id += 1
                
                elif element.tag == 'Properties':
                    run_callback(index, 'Level Properties', max)
                        
                    for el in element:
                        if el is etree.Comment:
                            continue
                        
                        if el.tag == 'Property':
                            this.properties[el.get('name')] = el.get('value')
                
                elif element.tag == 'Room':
                    run_callback(index, 'Room', max)
                        
                    for el in element:
                        if el is etree.Comment:
                            continue
                        
                        if el.tag == 'AbsoluteLocation':
                            this.room = tuple([float(_) for _ in el.get('value').split()])
                elif element.tag == 'Challenges':
                    for challenge in element:
                        challenge : etree.ElementBase
                        if challenge is etree.Comment:
                            continue
                        
                        if challenge.tag == 'Challenge':
                            this.challenges.append(this.Challenge(challenge))
                        
                
                elif callable(load_callback):
                    run_callback(index, element.tag, max)
                    
                index += 1
                
                run_callback(index, 'Done!', max)
                
            except:
                log_exception()
                if not ignore_errors:
                    raise
            
    def export(
        this,
        filename : str = None,
        exportObjects : bool = False,
    ) -> bytes:
        """Export level

        Args:
            filename (str, optional): Path to level. Defaults to Level.filename.
            exportObjects (bool, optional): Whether to export objects. Defaults to False.

        Raises:
            TypeError: Path is not a file.

        Returns:
            bytes: XML file.
        """
        if filename == None:
            if this.filename:
                filename = this.filename
        else:
            this.filename = filename
        
        xml : etree.ElementBase = etree.Element('Objects')
        for object in this.objects:
            if exportObjects:
                object.export()
            
            xml.append(object.getLevelXML())
        
        room = etree.Element('Room')
        etree.SubElement(room, 'AbsoluteLocation', value = ' '.join([str(_) for _ in this.room]))

        properties = etree.Element('Properties')
        for name in this.properties:
            value = this.properties[name]
            etree.SubElement(properties, 'Property', name = name, value = value)
        
        if len(properties):
            xml.append(properties)
        
        challenges : etree._Element = etree.Element('Challenges')
        
        for challenge in this.challenges:
            challenges.append(challenge.getXML())
        
        if len(challenges):
            xml.append(challenges)
        
        this.xml = xml
        
        output = etree.tostring(xml, pretty_print=True, xml_declaration=True, encoding='utf-8')
        
        if (file := this.filesystem.get(filename)) != None:
            if isinstance(file, File):
                file.write(output)
            else:
                raise TypeError(f'Path {filename} is not a file.')
                
        else:
            this.filesystem.add(filename, output)
        
        return output
    
    def addObject(
        this,
        filename : str | Object,
        properties : dict = {},
        pos : tuple[float,float] = (0,0),
        name : str = 'Obj'
    ):
        """Add object to level.

        Args:
            filename (str | Object): Filename for object. If it's a wmwpy.classes.Object class, then it will use that instead.
            properties (dict, optional): Object properties. Defaults to {}.
            pos (tuple[x,y], optional): Position of object in level. Defaults to (0,0).
            name (str, optional): Name of object. May get renamed if object with name alread exists. Defaults to 'Obj'.

        Returns:
            Object: wmwpy Object.
        """
        if not isinstance(filename, Object):
            filename = Object(
                filename,
                filesystem = this.filesystem,
                properties = properties,
                pos = pos,
                name = name,
            )
        else:
            filename.name = name
            filename.pos = pos
            filename.setProperty(properties)
        
        obj = filename
        
        id = 0
        while this.getObjectById(id) != None:
            id += 1
        
        if this.getObject(obj.name) != None:
            objnum = 0
            name = obj.name
            
            while this.getObject(obj.name) != None:
                objnum += 1
                obj.name = f'{name}{str(objnum)}'
        
        obj.id = id
        this.objects.append(obj)
        obj.scale = this.scale
        
        return obj
    
    def getObjectById(this, id : int) -> Object:
        """Get an Object by it's id

        Args:
            id (int): Object id to find

        Returns:
            Object: wmwpy Object
        """
        for obj in this.objects:
            if obj.id == id:
                return obj
        return None
    
    def getObject(this, name : str):
        """
        Get object by name

        Args:
            name (str): Object name.
        """
        
        for obj in this.objects:
            if obj.name == name:
                return obj
    
    class Challenge():
        def __init__(
            this,
            xml : etree.ElementBase = None,
            id : str = '',
            requirements : dict[str, dict[str, str]] = {},
        ) -> None:
            """A level challenge used in wmw2

            Args:
                xml (etree.Element, optional): The xml of the challenge. If it is `None`, it will just use the other values. Defaults to None.
                id (str, optional): The id of the challenge. Defaults to ''.
                requirements (dict[str, dict[str, str]], optional): The requirements as a `dict`. Defaults to {}.
            
            Requirements format:
                ```python
                dict[str, dict[str, str]]
                ```
                Example:
                ```python
                {
                  'WindWait' : {
                    'seconds' : '1'
                  },
                  'Duck' : {
                    'count' : '2'
                  }
                }
                ```
            """
            this.xml = xml
            this.id = id
            this.requirements : dict[str, dict[str, str]] = copy.deepcopy(requirements)
            
            if isinstance(this.xml, etree._Element):
                this.readXML()
        
        def readXML(this):
            """Read the XML of the challenge. If the XML wasn't set, it'll just return `None`
            """
            if not isinstance(this.xml, etree._Element):
                return
            
            this.id = this.xml.get('id', '')
            
            for element in this.xml:
                # so I can acess the attributes in vscode
                element : etree.ElementBase
                
                if element is etree.Comment:
                    continue
                
                requirement = copy.deepcopy(element.attrib)
                
                this.requirements[element.tag] = requirement
        
        def getXML(this):
            """Get the XML for the challenge.
            
            Returns:
                lxml.etree.Element: lxml etree Element.
            """
            root : etree._Element = etree.Element('Challenge', id = this.id)
            
            for name in this.requirements:
                requirement = this.requirements[name]
                
                etree.SubElement(root, name, **requirement)
            
            this.xml = root
            
            return root
            



from lxml import etree

def findTag(root : etree.ElementBase, tag : str) -> etree.ElementBase:
    """Find an element in an etree element.

    Args:
        root (etree.ElementBase): The root element to search.
        tag (str): The element to find.

    Returns:
        etree.ElementBase: etree Element
    """
    element = 0
    curTag = ''
    for e in root:
        # print(e.tag)
        if e.tag is etree.Comment:
            continue
        curTag = e.tag
        if curTag == tag:
            return e
        
    return None

def strbool(value : str) -> bool:
    """Convert a string to a bool

    Args:
        value (str): String. Can be 'true', 'false', 't', 'f', '0', or '1'. The string is case insensative.

    Returns:
        bool: Python `bool` object.
    """
    if isinstance(value, bool):
        return value
    else:
        return value.lower() in ("yes", "true", "t", "1")



__version__ = "0.3.0-beta"
__author__ = 'ego-lay-atman-bay'

import typing

from .game import Game
from . import classes
from . import Font
from .classes import widget
from . import Utils
from .gametemplate import GAMES
from .Utils import filesystem

def load(
    gamepath : str,
    platform : typing.Literal['android', 'ios'] = 'android',
    game : str = 'WMW',
    assets : str = None,
    db : str = None,
    profile : str = None,
    baseassets : str = None,
    hook : typing.Callable[[int, str, int], typing.Any] = None,
) -> Game:
    """load game

    Args:
        gamepath (str): Folder to extracted game.
        platform (Literal['android', 'ios'], optional): What platform this game is for. Can be 'android' or 'ios'. Defaults to 'android'.
        game (str, optional): Which game is being loaded. A full list of games is in the `gametemplate.GAMES` variable. Defaults to 'WMW'. 
        assets (str, optional): Relative path to assets folder. Defaults to '/assets'.
        db (str, optional): Relative path to database file from assets folder. Defaults to '/Data/water.db'.
        profile (str, optional): Relative path to profile file in WMW2. Defaults to `None`
        baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
        hook (Callable[[int, str, int], Any], optional): Hook for loading assets, useful for guis. The function gets called with the paramaters `(progress : int, current : str, max : int)`. Defaults to None.
    """
    
    game = game.upper()
    platform = platform.lower()
    
    platforms = {
        'android': {
            'assets': '/assets',
        },
        'ios': {
            'assets': '/Content',
        },
    }
    
    if assets == None:
        assets = platforms[platform]['assets']
    
    try:
        return GAMES[game](
            gamepath=gamepath,
            assets=assets,
            db=db,
            profile=profile,
            hook = hook,
            baseassets=baseassets
        )
    except:
        return Game(
            gamepath=gamepath,
            assets=assets,
            db=db,
            profile=profile,
            hook = hook,
            baseassets=baseassets
        )



import lxml
from lxml import etree
import io
from copy import deepcopy
from PIL import Image, ImageTk, ImageDraw
import numpy
import math

from ..gameobject import GameObject
from .sprite import Sprite
from ..Utils.filesystem import *
from ..Utils.rotate import rotate

from ..Utils.XMLTools import strbool
class Object(GameObject):
    def __init__(
        this,
        file : str | bytes | File,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None,
        assets : str = '/assets',
        baseassets : str = '/',
        properties : dict = {},
        pos : tuple | str = (0,0),
        name : str = 'Obj',
        scale : int = 10,
    ) -> None:
        """Get game object. Game object is `.hs` file.

        Args:
            file (str | bytes | File): Object file.
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
            assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
            properties (dict, optional): Object properties that override default properties. Defaults to {}.
            pos (tuple | str, optional): Position of object. Defaults to (0,0).
            name (str, optional): Name of object. Defaults to 'Obj'.
            scale (int, optional): The image scale. Defaults to 10.
        """
        
        super().__init__(filesystem, gamepath, assets, baseassets)
        
        this.file = super().get_file(file)
        
        this._properties = deepcopy(properties)
        if isinstance(pos, str):
            this.pos = tuple([float(a) for a in pos.split()])
        else:
            this.pos = tuple(pos)
        
        this.xml : etree.ElementBase = etree.parse(this.file).getroot()
        this.sprites : list[Sprite] = []
        this.shapes : list[Shape] = []
        this.UVs : list[tuple[int,int]] = []
        this.VertIndices : list[int] = []
        this.defaultProperties = {}
        this.properties = {}
        this.name = name
        this.size = (0,0)
        this.id = 0
        
        this._background : list[Sprite] = []
        this._foreground : list[Sprite] = []
        this._PhotoImage : dict[str, ImageTk.PhotoImage] = {}
        
        this._offset = [0,0]
        this.scale = scale
        
        this.readXML()
        
        if isinstance(file, File):
            this.filename = file.path
    
    def getOffset(this) -> tuple[float,float]:
        """Get the center offset for the Object image

        Returns:
            tuple[float,float]: (x,y)
        """
        rects = []
        this._background : list[Sprite] = []
        this._foreground : list[Sprite] = []
        
        for sprite in this.sprites:
            if 'visible' in sprite.properties:
                if not strbool(sprite.properties['visible']):
                    continue
            if ('isBackground' in sprite.properties) and strbool(sprite.properties['isBackground']):
                this._background.append(sprite)
            else:
                this._foreground.append(sprite)
                
            pos = numpy.array(sprite.pos)
            size = (numpy.array(sprite.image.size) / sprite.scale) * [1,-1]
            
            rects.append(
                tuple(pos - (size / 2))
            )
            rects.append(
                tuple(pos + (size / 2))
            )
            
        rects = numpy.array(rects).swapaxes(0,1)
        
        min = numpy.array([math.floor(v.min()) for v in rects])
        max = numpy.array([math.ceil(v.max()) for v in rects])
        
        
        this.size = max - min
        this._offset = [a.mean() for a in numpy.array([min,max]).swapaxes(0,1)]
        
        return this._offset
    
    @property
    def background(this) -> Image.Image:
        """The background image of this Object

        Returns:
            PIL.Image.Image: PIL Image
        """
        this.getOffset()
        
        image = Image.new('RGBA', tuple(this.size * this.scale), (0,0,0,0))
        
        for sprite in this._background:
            size = (numpy.array(sprite.image.size) / sprite.scale) * [1,-1]
            pos = this.truePos(
                sprite.pos,
                size,
                this.size,
                scale = this.scale,
                offset = this._offset
            )
            
            # print(f'{pos = }')
            
            image.alpha_composite(
                sprite.image,
                tuple([round(x) for x in pos]),
            )
        image = this.rotateImage(image)
        
        return image
    
    @property
    def background_PhotoImage(this) -> ImageTk.PhotoImage:
        """Tkinter PhotoImage of this Object

        Returns:
            ImageTk.PhotoImage: Tkinter PhotoImage
        """
        this._PhotoImage['background'] = ImageTk.PhotoImage(this.background)
        return this._PhotoImage['background']
    
    @property
    def foreground(this) -> Image.Image:
        """The foreground of the Object image

        Returns:
            PIL.Image.Image: PIL Image
        """
        this.getOffset()
        image = Image.new('RGBA', tuple(this.size * this.scale), (0,0,0,0))
        
        for sprite in this._foreground:
            size = (numpy.array(sprite.image.size) / sprite.scale) * [1,-1]
            pos = this.truePos(
                sprite.pos,
                size,
                this.size,
                scale = this.scale,
                offset = this._offset
            )
            
            # print(f'{pos = }')
            
            image.alpha_composite(
                sprite.image,
                tuple([round(x) for x in pos]),
            )
        image = this.rotateImage(image)
        
        return image
    
    @property
    def foreground_PhotoImage(this) -> ImageTk.PhotoImage:
        """Foregound Tkinter PhotoImage

        Returns:
            ImageTk.PhotoImage: Tkinter PhotoImage
        """
        this._PhotoImage['foreground'] = ImageTk.PhotoImage(this.foreground)
        return this._PhotoImage['foreground']
    
    @property
    def image(this) -> Image.Image:
        """Full Object image, with both the background and foreground.

        Returns:
            PIL.Image.Image: PIL Image
        """
        
        image = this.background
        image.alpha_composite(this.foreground)
        
        return image
    
    @property
    def offset(this) -> tuple[float,float]:
        """The center offset of the Object image

        Returns:
            tuple[float,float]: (x,y)
        """
        this.getOffset()
        offset = this._offset
        
        offset = this.rotatePoint(this._offset)
        
        return offset
    
    def rotatePoint(this, point : tuple = (0,0), angle : float = None) -> tuple[float,float]:
        """Rotate a point around (0,0)

        Args:
            point (tuple, optional): Point to rotate. Defaults to (0,0).
            angle (float, optional): Angle to rotate. Defaults to Object `Angle` property.

        Returns:
            tuple[float,float]: (x,y)
        """
        if angle == None:
            if 'Angle' in this.properties:
                angle = float(this.properties['Angle'])
            else:
                angle = 0
        
        if angle == 0:
            return point
        
        return rotate(point, degrees=-angle)
    
    def rotateImage(this, image : Image.Image) -> Image.Image:
        """Rotate an image the amount of degrees as the Object `Angle` property

        Args:
            image (PIL.Image.Image): Image to rotate

        Returns:
            PIL.Image.Image: Rotated PIL Image
        """
        if 'Angle' in this.properties:
            angle = float(this.properties['Angle'])
            image = image.rotate(angle, expand = True, resample = Image.BILINEAR)
        
        return image
    
    @property
    def PhotoImage(this) -> ImageTk.PhotoImage:
        """Tkinter PhotoImage of the Object image

        Returns:
            ImageTk.PhotoImage: Tkinter PhotoImage
        """
        this._PhotoImage['image'] = ImageTk.PhotoImage(this.image)
        return this._PhotoImage['image']
    
    @property
    def scale(this) -> int:
        """Object image scale
        """
        return this._scale
    @scale.setter
    def scale(this, value : int):
        this._scale = value
        for sprite in this.sprites:
            sprite.scale = this._scale
        
    def readXML(this):
        """Read object XML
        """
        
        # specifically specifying type so it's easier to use in vscode
        tags = {
            'Shapes': this._getShapes,
            'Sprites': this._getSprites,
            'UVs': this._getUVs,
            'VertIndices': this._getVertIndices,
            'DefaultProperties': this._getDefaultProperties,
        }
        
        for element in this.xml:
            if element is etree.Comment:
                continue
            if element.tag in tags:
                tags[element.tag](element)
        
        this.getProperties()
    
    def export(this, path : str = None) -> bytes:
        """Export object XML

        Args:
            path (str, optional): Filename for object. Defaults to Object.filename.

        Raises:
            TypeError: Path is not a file

        Returns:
            bytes: XML file.
        """
        xml : etree.ElementBase = etree.Element('InteractiveObject')
        
        shapes = etree.Element('Shapes')
        
        for shape in this.shapes:
            shape : Shape
            shapes.append(shape.getXML())
        
        if len(shapes) > 0:
            xml.append(shapes)
        
        sprites : etree.ElementBase = etree.Element('Sprites')
        
        for sprite in this.sprites:
            sprite : Sprite
            sprite.export()
            etree.SubElement(sprites, 'Sprite', sprite.properties)
        
        if len(sprites) > 0:
            xml.append(sprites)
        
        UVs : etree.ElementBase = etree.Element('UVs')
        
        for UV in this.UVs:
            pos = ' '.join([str(_) for _ in UV])
            etree.SubElement(UVs, 'UV', {'pos': pos})
        
        if len(UVs) > 0:
            xml.append(UVs)
        
        VertIndices : etree.ElementBase = etree.Element('VertIndices')
        
        for index in this.VertIndices:
            etree.SubElement(VertIndices, 'Vert', {'index': str(index)})
        
        if len(VertIndices) > 0:
            xml.append(VertIndices)
        
        DefaultProperties : etree.ElementBase = etree.Element('DefaultProperties')
        
        for name in this.defaultProperties:
            etree.SubElement(DefaultProperties, 'Property', {
                'name': name,
                'value': this.defaultProperties[name]
            })
        
        if len(DefaultProperties) > 0:
            xml.append(DefaultProperties)
        
        this.xml = xml
        output = etree.tostring(xml, pretty_print=True, xml_declaration=True, encoding='utf-8')
        
        if path == None:
            if this.filename:
                path = this.filename
        
        if path != None:
            if (file := this.filesystem.get(path)) != None:
                if isinstance(file, File):
                    file.write(output)
                else:
                    raise TypeError(f'Path {path} is not a file.')
                    
            else:
                this.filesystem.add(path, output)
        
        return output
    
    def updateProperties(this):
        """Update properties. Deletes any properties that are the same as defaultProperties.
        """
        properties = list(this.properties.keys())
        
        # for property in properties:
        #     if property in this.defaultProperties:
        #         if this.properties[property] == this.defaultProperties[property]:
        #             del this.properties[property]
        
        # if this.type != None:
        #     this.properties['Type'] = this.type

    def getLevelXML(
        this,
        filename : str = None,
    ) -> etree.ElementBase:
        """Gets XML to be used in levels.

        Args:
            filename (str, optional): Object filename. Defaults to Object.filename.

        Returns:
            etree.Element: lxml Element
        """
        if filename == None:
            if this.filename:
                filename = this.filename
        else:
            this.filename = filename
        
        xml = etree.Element('Object', name = this.name)
        etree.SubElement(xml, 'AbsoluteLocation', value = ' '.join([str(_) for _ in this.pos]))
        
        properties = etree.SubElement(xml, 'Properties')
        
        this.updateProperties()
        
        for name in this.properties:
            value = this.properties[name]
            
            etree.SubElement(properties, 'Property', name = name, value = str(value))
        
        this.getProperties()
        
        return xml
    
    @property
    def filename(this) -> str | None:
        """Object filename based on the `Filename` proprty
        """
        if 'Filename' in this.properties:
            return this.properties['Filename']
        else:
            return None
    @filename.setter
    def filename(this, value : str):
        this.properties['Filename'] = value
    
    @property
    def type(this) -> str | None:
        """The Object type, based off the `Type` property.
        """
        if 'Type' in this.properties:
            return this.properties['Type']
        elif 'Type' in this.defaultProperties:
            return this.defaultProperties['Type']
        else:
            return ''
    @type.setter
    def type(this, value : str):
        if not isinstance(value, str):
            raise TypeError('type is not a string')
        
        if not 'Type' in this.defaultProperties:
            this.defaultProperties['Type'] = value
        this.properties['Type'] = value
    
    def _getShapes(this, xml : etree.ElementBase):
        for element in xml:
            shape = Shape(element)
            this.shapes.append(shape)
        
    def _getSprites(this, xml : etree.ElementBase):
        for element in xml:
            if element is etree.Comment:
                continue
            
            if element.tag == 'Sprite':
                attributes = element.attrib
                sprite = Sprite(
                    file = this.filesystem.get(attributes['filename']),
                    filesystem = this.filesystem,
                    properties = attributes,
                    scale = this.scale,
                )
                this.sprites.append(sprite)
    
    def _getUVs(this, xml : etree.ElementBase):
        for element in xml:
            if element is etree.Comment:
                continue
            if element.tag == 'UV':
                pos = element.get('pos')
                this.UVs.append(tuple([float(_) for _ in pos.split()]))
    
    def _getVertIndices(this, xml : etree.ElementBase):
        for element in xml:
            if element is etree.Comment:
                continue
            if element.tag == 'Vert':
                index = element.get('index')
                this.VertIndices.append(int(index))
    
    def getProperties(this):
        for prop in this.defaultProperties:
            this.properties[prop] = this.defaultProperties[prop]
        for prop in this._properties:
            this.properties[prop] = this._properties[prop]
        return this.properties
    
    def _getDefaultProperties(this, xml : etree.ElementBase):
        for element in xml:
            if element is etree.Comment:
                continue
            if element.tag == 'Property':
                name = element.get('name')
                value = element.get('value')
                
                this.defaultProperties[name] = value
    
    def setProperty(this, property : str | dict, value : str = ''):
        """Set object property.

        Args:
            property (str | dict): Property name to set. If value is dict, it will combine the properties in the dict with the current properties.
            value (str, optional): Property value. Defaults to ''.
        """
        if isinstance(property, dict):
            for name in property:
                this.properties[name] = property[name]
            return
        this.properties[property] = value

class Shape(GameObject):
    def __init__(this, xml : etree.ElementBase = None) -> None:
        """Shape for Object

        Args:
            xml (etree.Element, optional): lxml Element. Defaults to None.
        """
        this.points : list[tuple[float,float]] = []
        this.xml = xml
        
        this.readXML()
    
    def readXML(this):
        """Read XML if any.
        """
        if this.xml == None:
            return
        for element in this.xml:
            if element is etree.Comment:
                continue
            if element.tag == 'Point':
                pos : str = element.get('pos')
                point = tuple([float(_) for _ in pos.split()])
                this.points.append(point)
    
    def getXML(this) -> etree.ElementBase:
        """Gets Shape XML for Object.

        Returns:
            etree.Element: lxml Element.
        """
        xml : etree.ElementBase = etree.Element('Shape')
        for point in this.points:
            etree.SubElement(xml, 'Point', {'pos': ' '.join([str(_) for _ in point])})
        this.xml = xml
        return xml
    
    @property
    def image(this) -> Image.Image:
        """Get the Shape image

        Returns:
            PIL.Image.Image: PIL Image
        """
        points = numpy.array(this.points).swapaxes(0,1)
        
        min = numpy.array([math.floor(v.min()) for v in points])
        max = numpy.array([math.ceil(v.max()) for v in points])
        
        offset = numpy.array([a.mean() for a in numpy.array([min,max]).swapaxes(0,1)])
        # offset = offset * [1,-1]
        # print(f'{offset = }')
        
        size = max - min
        
        image = Image.new('1', tuple([math.ceil(x) + 1 for x in size]), 1)
        draw = ImageDraw.Draw(image)
        
        # size = size * [1,-1]
        # print(f'{size = }')
        for n in range(len(this.points)):
            point = this.points[n]
            previous = (this.points[(n - 1) % len(this.points)])
            
            line = numpy.array([point, previous])
            # line = line * [1,-1]
            line = numpy.array(this.truePos(
                line,
                (1,1),
                size,
                offset,
            ))
            
            # print(line)
            
            line = line.flatten()
            line = tuple([round(x) for x in line])
            
            
            draw.line(line, fill=0, width=1)
        
        return image



import io
from lxml import etree

from .. import filesystem
from ...classes import *

class ImagelistFile(filesystem.Reader):
    MIME = 'text/imagelist'
    EXTENSION = 'imagelist'
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return extension == this.EXTENSION
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return Imagelist(rawdata.getvalue(), **kwargs)
    
class SpriteFile(filesystem.Reader):
    MIME = 'text/sprite'
    EXTENSION = 'sprite'
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return mime == this.MIME
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return Sprite()
    
class ObjectFile(filesystem.Reader):
    MIME = 'text/hs'
    EXTENSION = 'hs'
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return mime == this.MIME
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return Object()
    
class XMLFile(filesystem.Reader):
    MIME = 'text/xml'
    EXTENSION = 'xml'
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return extension == this.EXTENSION
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs) -> etree.ElementBase:
        return etree.parse(rawdata).getroot()
    
TYPES = [ImagelistFile(), XMLFile(), ObjectFile(), SpriteFile()]



import lxml
from lxml import etree
from PIL import Image, ImageTk
from .widget import Widget, register_widget

class WT_GROUP(Widget):
    """
        Group for widgets?
    """
    def __init__(this, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

register_widget('WT_GROUP', WT_GROUP,)



from lxml import etree
import io

# c++ code that can read curves collection files
# =================================================
# https://github.com/DannyTking/TalkingScene/blob/master/TalkingScene/WalaberEngine/Walaber/src/Math/CurveManager.cpp#L406
# https://github.com/DannyTking/TalkingScene/blob/master/TalkingScene/WalaberEngine/Walaber/include/Math/Curve.h#L19
# =================================================
# If anyone know how to read c++ code, please help.

class curvesCollection():
    MAGIC = 0xC081EC54
    MAGIC_V1 = 0xC081EC55
    
    def __init__(this, file : bytes) -> None:
        if isinstance(file, io.BytesIO):
            file = file.getvalue()
        elif not isinstance(file, bytes):
            raise TypeError('file must be bytes or BytesIO object.')
        
        this.rawdata = file
        
        int.from_bytes()
        
    def read(this):
        this.header = {
            'magic': None,
            'version': None,
            'curves': None,
            'flags': None,
            'groups': None,
        }
        
        magic = int.from_bytes(this.rawdata[0:4], byteorder='little')
        
        if magic == this.MAGIC:
            this.header['version'] = 1
        elif magic == this.MAGIC_V1:
            this.header['version'] = int(this.rawdata[5])
        else:
            raise ValueError('file not a curves collection .bin file')
        
        

if __name__ == "__main__":
    pass


from .object import Object
from .level import Level
from .layout import Layout
from .sprite import Sprite
from .imagelist import Imagelist
from .widget import Widget

__all__ = ['Level', 'Layout', 'Sprite', 'Object', 'Imagelist', 'Widget']



from PIL import ImageTk

from .widget import Widget, register_widget, WIDGETS, get_widget

from .WT_PUSH_BUTTON import WT_PUSH_BUTTON
from .WT_LABEL import WT_LABEL
from .WT_TOGGLE import WT_TOGGLE
from .WT_SLIDER import WT_SLIDER
from .WT_PROGRESS_BAR import WT_PROGRESS_BAR
from .WT_GROUP import WT_GROUP
from .WT_ICON_LIST import WT_ICON_LIST
from .WT_SCROLLABLE_CAMERA import WT_SCROLLABLE_CAMERA
from .WT_SCROLLABLE_SET import WT_SCROLLABLE_SET
from .WT_CANVAS import WT_CANVAS
from .WT_FINGER_CATCHER import WT_FINGER_CATCHER

__all__ = ['Widget', 'WIDGETS', 'register_widget', 'get_widget']



from ..Utils.textures import getHDFile, getTextueSettings, getTexture
from ..Utils.filesystem import *
from ..Utils import joinPath, Texture
from ..gameobject import GameObject

import numpy
import PIL.Image
from lxml import etree
from copy import deepcopy

import io
import os

class Imagelist(GameObject):
    TEMPLATE = b"""<?xml version="1.0"?>
    <ImageList imgSize="512 512" file="" textureBasePath="/Textures/">
    </ImageList>
    """
    
    class Type():
        IMAGELIST = 0
        PAGES = 1
    
    def __init__(
        this,
        file : str | bytes | File = None,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None, assets : str = '/assets',
        baseassets : str = '/',
        HD : bool = False,
    ) -> None:
        """
        Get imagelist from file
        Args:
            file (str | bytes | File): File to read.
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            HD (bool, optional): Whether to use HD textures. Defaults to False.
            gamepath (str, optional): Path to game. Only is used when filesystem is `Folder` or `None`. Defaults to None.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
            assets (str, optional): Path to assets relative to gamepath. Defaults to '/assets'.
        Raises:
            FileNotFoundError: Filesystem is not usable and no gamepath.
        """
        
        super().__init__(filesystem, gamepath, assets, baseassets)
        
        this.file = super().get_file(file, template = this.TEMPLATE)

        this.HD = HD
        this.xml : etree.ElementBase = etree.parse(this.file).getroot()
        this.textureBasePath : str = '/Textures/'
        this.pages : list[Imagelist.Page] = []
        this.type = this.Type.IMAGELIST
        this.filename = ''

        # this.images = {}

        this.read()
    
    def read(this):
        """Read the imagelist xml.
        """
        this.type = this.Type.IMAGELIST
        
        for element in this.xml:
            if element is etree.Comment:
                continue
            
            if element.tag == 'Page':
                this.type = this.Type.PAGES
                page = this.Page(
                    element,
                    filesystem = this.filesystem,
                    HD = this.HD,
                )
                
                this.pages.append(page)
        
        if this.type == this.Type.IMAGELIST:
            page = this.Page(
                this.xml,
                filesystem = this.filesystem,
                HD = this.HD,
            )
            
            this.pages.append(page)
    
    def export(
        this,
        path : str = None,
        exportImage : bool = False,
        imageFormat : str = 'webp',
        removeImageFiles : bool = False,
    ):
        """Export the xml of the imagelist.

        Args:
            path (str, optional): Path to the file in the filesystem to write to. If `None`, it will not save to a file, only report the output. Defaults to None.
            exportImage (bool, optional): Whether to also export the atlas image(s). If there are multiple pages, it'll append `_split_#` to the end of the filenames. Defaults to False.
            imageFormat (str, optional): What format to export the images as. Defaults to 'webp'.
            removeImageFiles (bool, optional): Remove image files from filesystem. Defaults to False.

        Raises:
            TypeError: Path is an existing folder.

        Returns:
            bytes: The xml output as bytes.
        """
        if path == None:
            if this.filename:
                path = this.filename
        else:
            this.filename = path
        
        if removeImageFiles:
            this.removeImageFiles()
        
        if path != None:
            if exportImage:
                if this.type == this.Type.PAGES:
                    index = 0
                    for page in this.pages:
                        index += 1
                        
                        filename = os.path.splitext(path)[0]
                        filename = f'{filename}_split_{str(index)}.{imageFormat}'
                        
                        page.file = filename
                        page.exportAtlas(filename = filename, format = imageFormat)
                        
                else:
                    page : this.Page = this.pages[0]
                    
                    filename = os.path.splitext(path)[0]
                    filename = f'{filename}.{imageFormat}'
                    
                    page.file = filename
                    page.exportAtlas(filename = filename, format = imageFormat)
        
        if this.type == this.Type.IMAGELIST:
            page = this.pages[0]
            xml = page.getXML(type = this.type)
        else:
            xml = etree.Element('Imagelist')
            for page in this.pages:
                xml.append(page.getXML(type = this.type))
            
        output = etree.tostring(xml, pretty_print=True, xml_declaration=True, encoding='utf-8')
        
        if path != None:
            if (file := this.filesystem.get(path)) != None:
                if isinstance(file, Folder):
                    raise TypeError(f'Path {path} is not a file.')
                
                file.write(output)
                
            else:
                file = this.filesystem.add(path, output)
        
        
        return output
    
    def combinePages(this):
        """Combine all the pages in this Imagelist into 1 Page
        """
        if this.type == this.Type.IMAGELIST:
            return
        
        main = this.pages[0]
        for i in range(len(this.pages) - 1):
            page : this.Page = this.pages[i + 1]
            for name in page.images:
                image = page.images[name]
                main.add(
                    image.name,
                    image.image,
                    image.properties,
                    replace = False
                )
                
        main.id = None
        main.exportAtlas()
        
        this.type = this.Type.IMAGELIST
        
        this.pages = [main]
                
    
    def getImage(this, name : str):
        """Get image from imagelist.

        Args:
            name (str): Name of image.

        Returns:
            Imagelist.Page.Image: Imagelist Image.
        """
        for page in this.pages:
            image = page.getImage(name)
            if image:
                return image
            
        return None
        
        # return this.filesystem.get(os.path.join(this.textureBasePath, name))
    
    def removeImageFiles(this):
        """Remove all image files in imagelist from filesystem.
        """
        for page in this.pages:
            page.removeImageFiles()
    
    class Page(GameObject):
        HD = False
        id = None
        imgSize = (1,1)
        textureBasePath = '/Textures/'
        file = ''

        def __init__(
            this,
            element : etree.ElementBase,
            filesystem: Filesystem | Folder = None,
            gamepath: str = None,
            assets: str = '/assets',
            HD = False,
        ) -> None:
            """Page for Imagelist. This is also used when imagelist is not in pages format.

            Args:
                element (etree.Element): lxml elment
                filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
                gamepath (str, optional): Gamepath used if filesystem is not specified. Defaults to None.
                assets (str, optional): Assets path relative to gamepath. Only used if filesystem is not specified. Defaults to '/assets'.
                HD (bool, optional): Use HD graphics. Defaults to False.
            """
            super().__init__(filesystem, gamepath, assets)
            
            this.HD : bool = HD
            
            this.xml : etree.ElementBase = element
            
            this.atlas = None
            this.images : dict[str, Imagelist.Page.Image] = {}
            
            this.read()
        
        class Image(GameObject):
            def __init__(
                this,
                atlas : PIL.Image.Image,
                properties : dict,
                textureBasePath = '/Textures',
                filesystem: Filesystem | Folder = None,
                gamepath: str = None,
                assets: str = '/assets'
            ) -> None:
                """Image for Imagelist

                Args:
                    atlas (PIL.Image.Image): Atlas file containing all images
                    properties (dict): Properties for Image
                    filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
                    gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
                    assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
                """
                super().__init__(filesystem, gamepath, assets)

                this.atlas = atlas
                this.properties = deepcopy(properties)

                this.size : tuple[int,int] = (1,1)
                this.offset : tuple[float,float] = (0,0)
                this.rect : tuple[int,int,int,int] = (0,0,0,0)
                this.name : str = ''
                this.textureBasePath : str = textureBasePath

                this.image = PIL.Image.new('RGBA', this.size)

                this.rawdata = io.BytesIO()

                this.getData()
                this.getImage()

            def getData(this):
                """Get properties from xml.
                """
                if 'size' in this.properties:
                    this.size = tuple([int(v) for v in this.properties['size'].split()])
                if 'offset' in this.properties:
                    this.offset = tuple([int(v) for v in this.properties['offset'].split()])
                if 'rect' in this.properties:
                    this.rect = tuple([int(v) for v in this.properties['rect'].split()])
                if 'name' in this.properties:
                    this.name = this.properties['name']

            def getImage(this):
                """Get image from atlas.

                Returns:
                    PIL.Image.Image: PIL Image.
                """
                this.image = this.atlas.crop(numpy.add(this.rect, (0,0) + this.rect[0:2]))
                this.image = this.image.resize(this.size)
                
                this.image.save(this.rawdata, format = os.path.splitext(this.name)[1][1::])
                return this.image

            def show(this):
                """Show image with default image viewer.
                """
                this.image.show()
            
            def updateProperties(this):
                """Update properties dict.
                """
                this.properties['name'] = this.name
                this.properties['size'] = ' '.join([str(x) for x in this.size])
                this.properties['rect'] = ' '.join([str(x) for x in this.rect])
                this.properties['offset'] = ' '.join([str(x) for x in this.offset])
            
            def getXML(this):
                """Get xml for image.

                Returns:
                    lxml.etree.Element: lxml element
                """
                this.updateProperties()
                xml : etree.ElementBase = etree.Element('Image', **this.properties)
                
                return xml
            
            def removeFile(this):
                """Remove file from filesystem.
                """
                return this.filesystem.remove(this.filename)
                
            
            def saveFile(this, replace : bool = False):
                """Save image to filesystem.

                Args:
                    replace (bool, optional): Whether to replace any existing file. Defaults to False.
                """
                this.image.save(this.rawdata, os.path.splitext(this.name)[1][1::])
                this.filesystem.add(
                    this.name,
                    file = this.rawdata.getvalue(),
                    replace = replace
                )
            
            
            @property
            def filename(this) -> str:
                """Image filepath in the Filesystem

                Returns:
                    str: Full filepath in the Filesystem
                """
                return this.filesystem.get(this.name).path

        
        def read(this):
            """Read xml.
            """
            this.properties = deepcopy(this.xml.attrib)
            if 'imgSize' in this.properties:
                this.size = tuple([int(v) for v in this.properties['imgSize'].split()])
            if 'textureBasePath' in this.properties:
                this.textureBasePath = this.properties['textureBasePath']
            if 'file' in this.properties:
                this.file = this.properties['file']
            if 'id' in this.properties:
                this.id = this.properties['id']
    
            if this.HD:
                hd = getHDFile(this.file)
                if this.filesystem.exists(hd):
                    this.file = hd
    
            this.fullAtlasPath = ''
    
            if this.gamepath:
                this.fullAtlasPath = joinPath(this.gamepath, this.assets, this.file)
                # print(this.fullAtlasPath)
    
            this.getAtlas()
            this.getImages()
    
        def getAtlas(this):
            """Get atlas image.
            """
            if not this.file:
                image = Texture(PIL.Image.new('RGBA', this.size))
                this.atlas = image.image.copy()
                return
            if this.filesystem.exists(this.file):
                file = this.filesystem.get(this.file)
                image = Texture(file.read())
                this.atlas = image.image.copy()
            else:
                this.textureSettings = getTextueSettings(
                    this.gamepath,
                    this.assets,
                    joinPath(os.path.dirname(os.path.dirname(this.textureBasePath)), 'Data/textureSettings.xml'),
                    this.name
                )
    
                this.atlas = getTexture(this.fullAtlasPath, this.textureSettings, this.size)
    
        def getImages(this):
            """Get images from xml.
            """
            for image in this.xml:
                if not image.tag is etree.Comment:
                    if image.tag == 'Image':
                        texture = this.Image(
                            this.atlas,
                            properties = image.attrib,
                            textureBasePath = this.textureBasePath,
                            filesystem = this.filesystem.get(this.textureBasePath),
                        )
                        this.images[image.get('name')] = texture
                        
                        # print(f'{this.textureBasePath = }')
                        # print(f'{texture.name = }')
    
                        this.filesystem.add(
                            joinPath(this.textureBasePath, texture.name),
                            texture.rawdata.getvalue(),
                            replace = True
                        )
    
        def getImage(this, name : str) -> Image:
            """Get an image from the imagelist

            Args:
                name (str): Name of image.

            Returns:
                Imagelist.Page.Image: Imagelist Image.
            """
            if name in this.images:
                return this.images[name]
            else:
                return None
        
        def add(this,
                name : str,
                image : PIL.Image.Image,
                properties : dict = {},
                replace = False
            ):
            """Add image to imagelist.

            Args:
                name (str): Name of image file used in-game.
                image (PIL.Image.Image): Image to use.
                properties (dict, optional): Additional properties for image. Defaults to {}.
                replace (bool, optional): Whether to replace existing image if there is a conflict. Defaults to False.

            Raises:
                NameError: Image already exists.
            """
            if name in this.images:
                # print(f'Warning: "{name}" already in imagelist.')
                if not replace:
                    raise NameError(f'Image "{name}" already exists.')
            properties['name'] = name
            properties['rect'] = ' '.join([str(_) for _ in (0,0) + image.size])
            
            texture = this.Image(
                image,
                properties
            )
            this.images[name] = texture
            
            this._getrect()
        
        def exportAtlas(this, filename = None, gap : tuple = (1,1), format : str = 'webp', ):
            """Export the atlas image into the Filesystem. This function recreates the imagelist, so you need to also export the xml using `getXML()`.

            Args:
                gap (tuple, optional): Gap between each image. Defaults to (1,1).
                filename (str, optional): Filename of image. Defaults to `file` property.
                format (str, optional): Format to save image as. Defaults to 'webp'.

            Returns:
                PIL.Image.Image: PIL Image.
            """
            this._getrect(gap = gap)
            this._updateAtlas()
            file = io.BytesIO()
            
            this.atlas.save(file, format=format)
            
            if filename == None:
                filename = f'{os.path.splitext(this.file)[0]}.{format}'
            
            this.file = filename
            
            if this.filesystem.exists(filename):
                this.filesystem.get(filename).rawdata = file
            else:
                this.filesystem.add(filename, file.getvalue())
            
            return this.atlas
        
        def updateProperties(this):
            """Updates all the properties dict.
            """
            this.properties['textureBasePath'] = this.textureBasePath
            this.properties['imgSize'] = ' '.join([str(n) for n in this.size])
            this.properties['file'] = this.file
            # print(this.properties['file'])
            if this.id == None:
                if 'id' in this.properties:
                   del this.properties['id']
            else:
                this.properties['id'] = str(this.id)
                
        
        def getXML(this, filename = None, type : int = 1):
            """Generates the xml for the page / imagelist.

            Args:
                filename (str, optional): Name of image. Defaults to file property.
                type (int, optional): Type of file. 0 for `Page`, 1 for `Imagelist`. Defaults to 1.

            Returns:
                lxml.etree.Element: lxml Element.
            """
            if filename != None:
                this.file = filename
            
            tag = 'Page' if type else 'Imagelist'
            
            this.updateProperties()
            xml : etree.ElementBase = etree.Element(tag, **this.properties)
            
            for name in this.images:
                image : this.Image = this.images[name]
                xml.append(image.getXML())
            
            this.xml = xml
            return this.xml
        
        def removeImageFiles(this):
            """Remove all image files from filesystem.
            """
            for name in this.images:
                this.images[name].removeFile()
            
        
        def _getrect(this, gap : tuple = (1,1)):
            """Update the rect for all images.

            Args:
                gap (tuple, optional): Gap between images. Defaults to (1,1).
            """
            x, y = gap
            maxheight = maxwidth = 0
            row = column = 0
            
            for image in this.images:
                image = this.images[image]
                image.rect = (x,y) + image.size
                
                if x > maxwidth:
                    maxwidth = x
                
                x += image.size[0] + gap[0]
                
                column += 1
                if x > this.size[0]:
                    x = gap[0]
                    y += maxheight + gap[1]
                    maxheight = 0
                    column = 0
                    row += 1
                
                if column == 0:
                    image.rect = (x,y) + image.size
                    x += image.size[0] + gap[0]
                
                if image.size[1] > maxheight:
                    maxheight = image.size[1]
            
            y += maxheight + gap[1]
            
            if y > this.size[1]:
                this.size = (this.size[0], y)
            
                
        def _updateAtlas(this):
            atlas : PIL.Image.Image = PIL.Image.new('RGBA', this.size)
            
            for image in this.images:
                image = this.images[image]
                image.atlas = atlas
                atlas.paste(image.image, image.rect[0:2])
            
            this.atlas = atlas
            return this.atlas
    
        def getNO_TEX(this):
            """## NEED TO UPDATE
            """
            NO_TEX_settings = getTextueSettings(
                this.gamepath,
                this.assets,
                joinPath(os.path.dirname(os.path.dirname(this.textureBasePath)), 'Data/textureSettings.xml'),
                os.path.join(this.textureBasePath, 'NO_TEX.png')
            )
            NO_TEX_image = PIL.Image.open(joinPath(this.gamepath, this.assets, this.textureBasePath, 'NO_TEX.png')).convert('RGBA')
            this.NO_TEX = this.Image(NO_TEX_image, {
                'size': ' '.join([str(x) for x in NO_TEX_image.size]),
                'rect': ' '.join([str(x) for x in (0,0) + NO_TEX_image.size]),
                'name': 'NO_TEX.png',
            })
            # this.Image(this.atlas, image.attrib)
    
    




import pathlib
import os
import io
from filetype import filetype
from PIL import Image
import zipfile
import natsort
import typing
import fnmatch

from .path import joinPath
# from . import Waltex
# from . import ImageUtils
# from ..classes import sprite
# from ..classes import Object

__all__ = ['Filesystem', 'File', 'Folder', 'FileBase']


FILE_READERS = []

# Filesystem object.

class Filesystem():
    def __init__(
        this,
        gamepath : str,
        assets : str,
        baseassets : str = '/'
    ) -> None:
        """wmwpy filesystem

        Args:
            gamepath (str): Path to game directory.
            assets (str): Path to assets folder relative to gamepath.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
        """
        this.gamepath = gamepath
        this.assets = assets
        this.baseassets = baseassets
        this.parent = this
        this.root = Folder(this, '/')
    
    def get(this, path : str):
        """Get file in filesystem.

        Args:
            path (str): Path to file or folder.

        Returns:
            File or Folder: File or Folder object.
        """
        return this.root.get(path)
    
    def add(this, path : str, file : str | bytes, replace = False):
        """
        Adds file with path to Folder.

        Args:
            path (str): Path to destenation of file. Must be str.
            file (str | bytes): File. This can be a path to the file to use, raw `bytes` of the file, or file-like object.
            replace (bool, optional): Replace any conflicting files? Defaults to False.

        Raises:
            TypeError: If file is not `str`, `bytes` or file-like object.

        Returns:
            File: File object of the added file.
        """
        # if isinstance(file, str):
        #     with open(file, 'rb') as f:
        #         file = f.read()
        #     # print('file')
        # elif isinstance(file, bytes):
        #     # print('bytes')
        #     pass
        # elif hasattr(file, 'read'):
        #     file = file.read()
        #     if isinstance(file, str):
        #         file = file.encode()
        #     # print('file-like')
        # else:
        #     raise TypeError(f"file can only 'str', 'bytes', or file-like object.")
        
        return this.root.add(path = path, content = file, replace = replace)
    
    def remove(this, path : str):
        """Remove a file or folder

        Args:
            path (str): Path to file or folder to remove.
        """
        return this.root.remove(path)
    
    def getAssets(
        this,
        extract_zip = False,
        split_imagelist = False,
        hook : typing.Callable[[int, str, int], typing.Any] = None
    ):
        """Scans the assets directory and loads all files into the filesystem. This is so wmwpy can modify files without modifying the actual files.

        Args:
            extract_zip (bool, optional): Extrack zip files? Defaults to False.
            split_imagelist (bool, optional): Split imagelist files? Defaults to False.
            hook (Callable[[int, str, int], Any], optional): Hook for loading assets, useful for guis. The function gets called with the paramaters `(progress : int, current : str, max : int)`. Defaults to None.

        Raises:
            FileNotFoundError: Assets folder does not exist.

        Returns:
            this: Current Filesystem object.
        """
                    
        def dirlength(path : str):
            count = 0
            for file in os.scandir(path):
                if file.is_dir():
                    count += dirlength(file.path)
                else:
                    count += 1
            return count

        
        # print(this.gamepath)
        # print(f'{this.gamepath = }\n{this.assets = }')
        
        assets = pathlib.Path(joinPath(this.gamepath, this.assets))
        
        if not assets.exists():
            raise FileNotFoundError(f'Folder {assets} does not exist')
        
        total = dirlength(assets.resolve().as_posix())
        current = 0
        
        for dir, subdir, files in os.walk(assets):
            for file in files:
                path = pathlib.Path('/', os.path.relpath(os.path.join(dir, file), assets)).as_posix()
                # print(path)
                
                if hook:
                    current += 1
                    hook(current, path, total)
                
                fileobj : File = this.add(path, os.path.join(dir, file))
                
                if fileobj.extension == 'zip' and extract_zip:
                    fileobj.read()
        
        return this
    
    def exists(this, fp : str) -> bool:
        """Test if file path exists.

        Args:
            fp (str): File path.

        Returns:
            bool: Whether the path exists.
        """
        return this.root.exists(fp)
    
    def listdir(this, path = '/', recursive = False, search = r'*') -> list:
        """Returns a list of files and subfolders in path.

        Args:
            path (str, optional): Path to folder to list. Defaults to '/'.
            recursive (bool, optional): Whether to include subfolders. Defaults to False.

        Returns:
            list: List of files and subfolders.
        """
        return this.get(path).listdir(recursive = recursive, search = search)
    
    def dump(this, output : str = None):
        """Dump the contents of the filesystem to the specified directory

        Args:
            output (str, optional): Path to output directory. Defaults to original path.
        """
        if output == None:
            output = joinPath(this.gamepath, this.assets)
        
        # print(f'output: {output}')
        
        files = this.listdir(recursive=True)
        for path in files:
            file = this.get(path)
            if file.is_dir():
                continue
            parts = pathlib.Path(path).parts
            
            if parts[0] in ['/', '\\', '']:
                parts = parts[1::]
            
            # print(f'new: {parts}')
            
            newpath = pathlib.Path(output, *parts)
            
            # print(f'writing: {newpath.as_posix()}')
            
            newpath.parent.mkdir(exist_ok=True)
            
            data = file.rawdata.getvalue()
            
            with open(newpath, 'wb') as f:
                f.write(data)
        
        
    
# Filesystem helpers
class FileBase():
    name = ''
    
    def __init__(this, parent, path : str):
        """File Base

        Args:
            parent (Folder): Parent. Use `None` for root.
            path (str): File path.
        """
        this._type = this._Type(None)
        this.name = pathlib.Path(path).parts[0]
        this.parent : Folder = parent
    
    def get(this, path : str):
        """Get file in filesystem.

        Args:
            path (str): Path to file or folder.

        Returns:
            File or Folder: File or Folder object.
        """
        path = pathlib.Path(path).as_posix()
        if path == '.':
            return this
        return this.parent.get(path)
    
    def remove(this, path : str = '.'):
        """Remove file or folder.

        Args:
            path (str, optional): Path to file or folder. If '.', it removes itself. Defaults to '.'.
        """
        path = pathlib.Path(path).as_posix()
        
        file = this.get(path)
        if file == None:
            return

        file.detatch()
    
    def detatch(self):
        """Detatch this file or folder from it's parent.
        """
        if self.parent == None:
            return
        self.parent.files.remove(self)
        self.parent = None
    
    def is_dir(this) -> bool:
        """Check whether this is a folder.

        Returns:
            bool
        """
        return this._type.value == this._Type.FOLDER
        
    @property
    def path(this) -> str:
        """Path to this file or folder from the root.

        Returns:
            str: Path this file or folder.
        """
        if this.parent == None or isinstance(this.parent, Filesystem):
            if isinstance(this, File) and not isinstance(this.parent, Filesystem):
                return this.name
            return '/'
        return pathlib.Path(this.parent.path, this.name).as_posix()
    
    @property
    def root(this) -> 'Folder':
        """Root Folder for this File or Folder.

        Returns:
            Folder: Root Folder for this File or Folder
        """
        if this.parent == None or this.name == '':
            return this
        return this.parent.root
    
    @property
    def filesystem(this) -> Filesystem:
        """Returns the Filesystem or root Folder for this File or Folder.

        Returns:
            Filesystem | Folder: Root Filesystem or Folder
        """
        if isinstance(this.parent, Filesystem):
            return this.parent
        if this.parent == None or this.name == '':
            return this
        return this.parent.filesystem
        
    class _Type():
        FOLDER = 0
        FILE = 1
        
        def __init__(this, type : int) -> None:
            this.value = type    
    
    
class File(FileBase):
    def __init__(
        this,
        parent,
        path: str,
        data : bytes | str | io.BytesIO,
    ):
        """File

        Args:
            parent (Folder): Parent. Use `None` for root.
            path (str): File path.
            content (bytes): Contents of file as bytes.
        """
        super().__init__(parent, path)
        this._type.value = this._Type.FILE
        
        this._datatype = 'raw'
        this._original_filename = ''
        
        if isinstance(data, bytes):
            this._rawcontent = data
        elif isinstance(data, str):
            if os.path.exists(data):
                this._rawcontent = data
                this._datatype = 'path'
                this._original_filename = data
            else:
                this._rawcontent = data.encode()
        elif isinstance(data, io.BytesIO):
            this._rawcontent = data.getvalue()
        elif isinstance(data, File):
            this._rawcontent = data._rawcontent
            this._datatype = data._datatype
            this._original_filename = data._original_filename
        elif hasattr(data, 'read'):
            data.seek(0)
            if hasattr(data, 'name'):
                this._original_filename = data.name
            
            data.seek(0)
            
            this._rawcontent = data.read()
            if isinstance(this._rawcontent, str):
                this._rawcontent = this._rawcontent.encode()
        else:
            this._rawcontent = bytes(data)
        
        this.content = None
        
        if this._datatype == 'raw':
            this._getdata()
        
    def _getdata(this):
        """Get data from file path
        """
        if this._datatype == 'path':
            with open(this._rawcontent, 'rb') as file:
                this._rawcontent = file.read()
            this._datatype = 'raw'
        
        this.rawdata = io.BytesIO(this._rawcontent)
        # seek back to the start to be able to read the data later.
        this.rawdata.seek(0)
        
        this.testFile()
        
    def testFile(this):
        """Tests what type of file this is."""
        this.type = filetype.guess(this.rawdata.getvalue())
        
        if this.type == None:
            this.type = None
            this.extension = os.path.splitext(this.name)[1][1::]
            if not this.extension:
                this.mime = f'text/raw'
            else:
                this.mime = f'text/{this.extension}'
            
        else:
            this.mime = this.type.mime
            this.extension = this.type.extension
            
        return this.mime
    
    @property
    def rawdata(this) -> io.BytesIO:
        """Returns the raw data of the file as a `BytesIO` object.

        Returns:
            io.BytesIO: Raw data of file.
        """
        if this._datatype == 'path':
            this._getdata()
        
        return this._rawdata
    @rawdata.setter
    def rawdata(this, value : io.BytesIO):
        this._rawdata : io.BytesIO = value
    
    @property
    def extension(this) -> str:
        """The file extension

        Returns:
            str: File extension
        """
        if this._datatype == 'path':
            return os.path.splitext(this._rawcontent)[1::]
        else:
            this._extension
    @extension.setter
    def extension(this, value):
        this._extension = value
        
    def read(this, **kwargs):
        """Read file.

        Returns:
            Any: Object for file.
        """
        
        if this._datatype == 'path':
            this._getdata()
        
        this.rawdata.seek(0)
        
        reader = Reader()
        
        for r in FILE_READERS:
            # print(r)
            if r.check(this.mime, this.extension, this.rawdata, filesystem = this.filesystem, **kwargs):
                reader = r
                break
        
        this.content = reader.read(this.mime, this.extension, this.rawdata, **kwargs)
        
        # if this.mime == 'image/waltex':
        #     this.content = Waltex(this.rawcontent.getvalue())
        #     this.image = this.content.image
        
        # elif this.mime.startswith('image/'):
        #     this.content = Image.open(this.rawcontent.getvalue())
        #     this.image = this.content
        
        # elif this.extension == 'zip':
        #     this.content = zipfile.ZipFile(this.rawcontent)
        #     if 'extract' in kwargs and kwargs['extract']:
        #         print(f'extracting {this.name}')
                
        #         files = this.content.namelist()
        #         for f in files:
        #             print(f)
                    
        #             content = this.content.read(f)
        #             this.root.add(f, content, replace = True)
                
        # elif this.mime.startswith('text/'):
        #     if this.extension == 'imagelist':
        #         this.content = ImageUtils.Imagelist(this.rawcontent.getvalue(), this.root, **kwargs)
        #         # I need to make Imagelist() accept a Folder or Filesystem, and raw file data.
        #         # raise NotImplementedError('Imagelist reading is currently not implemented yet.')
        #         # this.content = ImageUtils.Imagelist()
        #     else:
        #         this.content = this.rawcontent.read().decode(encoding=kwargs['encoding'])
        
        this.rawdata.seek(0)
        
        return this.content
    
    def add(this, path : str, file : bytes, replace = False):
        """Add file to folder.

        Args:
            path (str): New file path.
            content (bytes): Content of file in bytes.
            replace (bool, optional): Whether to replace any conflicting file.. Defaults to False.

        Raises:
            NotADirectoryError: Path to file contains file, not folder.
            FileExistsError: File already exists.

        Returns:
            File: Newly added File.
        """
        return this.parent.add(path, file, replace = replace)
    
    def exists(this, path : str):
        """Tests whether path exists.

        Args:
            path (str): Path to check.

        Returns:
            bool: Does path exist?
        """
        return this.parent.exists(path)
    
    def write(this, data : bytes) -> int:
        """Write data to file.

        Args:
            data (bytes): New data.

        Returns:
            int: bytes written.
        """
        this.rawdata.truncate(0)
        this.rawdata.seek(0)
        return this.rawdata.write(data)
    
    def listdir(this, recursive = False, search = r'*'):
        """Returns a list of files and subfolders in path.

        Args:
            path (str, optional): Path to folder to list. Defaults to '/'.
            recursive (bool, optional): Whether to include subfolders. Defaults to False.

        Returns:
            list: List of files and subfolders.
        """
        return this.parent.listdir(recursive = recursive, search = search)
    

class Folder(FileBase):
    def __init__(this, parent = None, path: str = None):
        """Folder

        Args:
            parent (Folder): Parent. Use `None` for root.
            path (str): Folder path.
        """
        if isinstance(parent, str) and path == None:
            path = parent
            parent = None
            
        if not path:
            path = '/'
        
        super().__init__(parent, path)
        this._type.value = this._Type.FOLDER
        this.files = []
        
    def add(this, path : str, content : bytes, replace = False) -> File:
        """Add file to folder.

        Args:
            path (str): New file path.
            content (bytes): Content of file in bytes.
            replace (bool, optional): Whether to replace any conflicting file.. Defaults to False.

        Raises:
            NotADirectoryError: Path to file contains file, not folder.
            FileExistsError: File already exists.

        Returns:
            File: Newly added File.
        """
        parts = pathlib.Path(path).parts
        
        file = this._getPath(pathlib.Path(*parts).as_posix())
        if len(parts) > 1:
            if file == None:
                file = Folder(this, parts[0])
                this.files.append(file)
                
            if file._type.value != file._Type.FOLDER:
                raise NotADirectoryError(f"{file.path} is not a directory.")
            
            return file.add(pathlib.Path(*parts[1::]).as_posix(), content, replace=replace)
        else:
            if file != None:
                if  not replace:
                    raise FileExistsError(f'File {file.path} already exists.')
                # print(f'File {file.path} already exists. Now replacing it.')
                this.files.remove(file)
            
            file = File(this, parts[0], data = content)
            this.files.append(file)
            
            return file
        
    def _getPath(this, path : str):
        """Get File or Folder with this path.

        Args:
            path (str): Path to File or Folder.

        Returns:
            File or Folder: File or Folder.
        """
        parts = pathlib.Path(path).parts
        file = None
        if parts[0] == '\\':
            file = this.root
        elif parts[0] == '..':
            file = this.parent
        else:
            for f in this.files:
                if f.name == parts[0]:
                    file = f
                    break
        return file
    
    def get(this, path : str) -> File:
        """Get file in filesystem.

        Args:
            path (str): Path to file or folder.

        Returns:
            File or Folder: File or Folder object.
        """
        if path == None:
            return None
        
        parts = pathlib.Path(path).parts
        if len(parts) == 0:
            return this
        
        file = this._getPath(path)
        if file == None:
            return file
        if file._type.value == this._Type.FOLDER:
            file = file.get(pathlib.Path(*parts[1::]))
        return file
    
    def exists(this, path : str) -> bool:
        """Tests whether path exists.

        Args:
            path (str): Path to check.

        Returns:
            bool: Does path exist?
        """
        return this.get(path) != None
    
    def listdir(this, recursive = False, search = r'*'):
        """Returns a list of files and subfolders in path.

        Args:
            path (str, optional): Path to folder to list. Defaults to '/'.
            recursive (bool, optional): Whether to include subfolders. Defaults to False.

        Returns:
            list: List of files and subfolders.
        """
        files = []
        for file in this.files:
            files.append(file.path)
            if recursive and file._type.value == this._Type.FOLDER:
                files = files + file.listdir(recursive = recursive)
        
        files = natsort.natsorted(files)
        files = fnmatch.filter(files, search)
        
        return files

class Reader():
    MIME = 'text/'
    EXTENSION = 'txt'
    
    def __init__(self, ) -> None:
        pass
    
    def check(this, mime : str, extension : str, rawdata : io.BytesIO, **kwargs):
        """Check file type.

        Args:
            mime (str): File mime.
            extension (str): File extension
            rawdata (io.BytesIO): Contents of file as file-like object.

        Returns:
            bool: Whether the file is of this type.
        """
        return mime.startswith(this.MIME)
    
    def read(this, mime : str, extension : str, rawdata : io.BytesIO, **kwargs):
        """Read file.

        Args:
            mime (str): Mime of file.
            extension (str): File extension.
            rawdata (io.BytesIO): File data as file-like object.

        Returns:
            Any: File data.
        """
        if 'endoding' in kwargs:
            encoding = kwargs['encoding']
        else:
            encoding = 'utf-8'
        
        return rawdata.getvalue().decode(encoding=encoding)

def register_reader(reader : Reader):
    """
    Add reader for file type.
    reader must inherit from `Utils.filesystem.Reader` class.

    Args:
        reader (Reader): Inherited `Reader` class.
    """
    
    if isinstance(reader, Reader):
        FILE_READERS.append(reader)
    else:
        raise TypeError('reader must inherit from `Utils.filesystem.Reader`')



import numpy


def rotate(point : tuple[float, float], degrees : float = 0):
    """Rotate a point around (0,0)

    Args:
        point (tuple[float,float]): Point to rotate.
        degrees (float, optional): Amount of degrees to rotate. Defaults to 0.

    Returns:
        tuple[float,float]: New point (x,y)
    """
    angle = numpy.deg2rad(degrees)
    
    point = numpy.array(point)
    # origin = numpy.array(origin)
    
    shape = point.shape
    
    if len(shape) > 1 and shape[1] >= 2:
        return numpy.array([rotate(p, degrees = degrees) for p in point])
    
    R = numpy.array([[numpy.cos(angle), -numpy.sin(angle)],
                  [numpy.sin(angle),  numpy.cos(angle)]])
    # o = numpy.atleast_2d(origin)
    m = numpy.dot(R, point)
    return float(m.T[0]), float(m.T[1])



import os
import typing

from .Utils.filesystem import *
from .Utils import Texture
from .classes import *

class Game():
    _DB = '/Data/water.db'
    _BASEASSETS = '/'
    _PROFILE = None
    
    game = 'WMW'
    
    def __init__(
        this,
        gamepath : str, assets : str = '/assets',
        db : str = '/Data/water.db',
        profile : str = None,
        baseassets : str = '/',
        platform : typing.Literal['android', 'ios'] = 'android',
        hook : typing.Callable[[int, str, int], typing.Any] = None,
    ) -> None:
        """load game

        Args:
            gamepath (str): Folder to extracted game.
            assets (str, optional): Relative path to assets folder. Defaults to '/assets'.
            db (str, optional): Relative path to database file from assets folder. Defaults to '/Data/water.db'.
            profile (str, optional): Relative path to profile file in WMW2. Defaults to `None`
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
            platform (Literal['android', 'ios'], optional): What platform this game is for. Can be 'android' or 'ios'. Defaults to 'android'.
            hook (Callable[[int, str, int], Any], optional): Hook for loading assets, useful for guis. The function gets called with the paramaters `(progress : int, current : str, max : int)`. Defaults to None.
        """
        if gamepath == None:
            return
        
        this.gamepath = os.path.abspath(gamepath)
        # print(f'{gamepath = }\n{this.gamepath = }')
        this.assets = assets
        this.db = db or this._DB
        this.profile = profile or this._PROFILE
        this.baseassets = baseassets or this._BASEASSETS
        this.platform = platform
        
        
        this.updateFilesystem(hook = hook)
        
    def updateFilesystem(this, hook : typing.Callable[[int, str, int], typing.Any] = None):
        this.filesystem = Filesystem(this.gamepath, this.assets)
        this.filesystem.getAssets(hook = hook)
        
    def Level(
        this,
        xmlPath : str = None,
        imagePath : str = None,
        load_callback : typing.Callable[[int, str, int], typing.Any] = None,
        ignore_errors : bool = False,
    ):
        """
        Load level

        Args:
            xmlPath (str, optional): Path to xml file. Defaults to None.
            imagePath (str, optional): Path to image file. Defaults to None.
        """
        if isinstance(xmlPath, File):
            xml = xmlPath
        else:
            xml = None
            if xmlPath:
                split = os.path.splitext(xmlPath)
                if split[1] == '':
                    if imagePath in ['', None]:
                        imagePath = '.'.join([split[0], 'png'])
                    xmlPath = '.'.join([split[0], 'xml'])
                
                xml = this.filesystem.get(xmlPath)
        
        if isinstance(imagePath, File):
            image = imagePath
        else:
            image = None
            if imagePath:
                image = this.filesystem.get(imagePath)
        
        level = Level(
            xml = xml,
            image = image,
            filesystem = this.filesystem,
            load_callback = load_callback,
            ignore_errors = ignore_errors,
        )
        if isinstance(xmlPath, File):
            level.filename = xmlPath.path
        else:
            level.filename = xmlPath
        
        return level
    
    def Object(
        this,
        object : str,
        **kwargs
    ):
        """
        Loads object

        Args:
            object (str): Path to `.hs` object file.

        Returns:
            classes.object.Object: Where's My Water? object.
        """
        
        
        if not isinstance(object, File):
            object = this.filesystem.get(object)
        
        obj = Object(
            object,
            filesystem = this.filesystem,
            **kwargs
        )
        if isinstance(object, File):
            obj.filename = object.path
        else:
            obj.filename = object
            
        
        return obj
    
    def Imagelist(
        this,
        imagelist : str = None,
        HD = False,
    ):
        """
        Load imagelist

        Args:
            imagelist (str): Path to `.imagelist` file. Defaults to None
            HD (bool, optional): Whether to use HD textures. Defaults to False.

        Returns:
            classes.imagelist.Imagelist: Imagelist object.
        """
        
        if not isinstance(imagelist, File):
            imagelist = this.filesystem.get(imagelist)
        
        imagelistObject = Imagelist(
            imagelist,
            filesystem = this.filesystem,
            HD = False,
        )
        if isinstance(imagelist, File):
            imagelistObject.filename = imagelist.path
        else:
            imagelistObject.filename = imagelist
        
        return imagelistObject
    
    def Sprite(
        this,
        sprite : str,
        **kwargs
    ):
        """
        Loads sprite.

        Args:
            sprite (str): Path to `.sprite` file.`

        Returns:
            classes.sprite.Sprite: Sprite object.
        """
        
        if not isinstance(sprite, File):
            sprite = this.filesystem.get(sprite)
        
        spriteObject = Sprite(
            sprite,
            filesystem = this.filesystem,
            **kwargs
        )
        if isinstance(sprite, File):
            spriteObject.filename = sprite.path
        else:
            spriteObject.filename = sprite.path
        
        return spriteObject
    
    def Texture(
        this,
        texture : str,
    ):
        """
        Get image texture. Doesn't matter if it's a `.waltex` image or not.

        Args:
            texture (str): Path to image file.

        Returns:
            Utils.textures.Texture: Texture object.
        """
        
        if not isinstance(texture, File):
            texture = this.filesystem.get(texture)
        
        return Texture(
            texture
        )
    
    def Layout(this, layout : str):
        raise NotImplementedError('load layout is not implemented yet.')
    
    def generateFileManifest(this, writeFile : bool = True, filename : str = '/FileManifest.txt'):
        """Generate the `FileManifest.txt` file needed for some games, such as WMM. This just generates a text file with the paths to every file in the `assets` folder (which includes the `FileManifest.txt` file).

        Args:
            writeFile (bool, optional): Write the manifest to the `FileManifest.txt` file. Defaults to True.
            filename (str, optional): Filename for FileManifest.txt. Defaults to 'filename'.

        Returns:
            str: Contents of `FileManifest.txt`
        """
        manifest = '\n'.join(this.filesystem.listdir(recursive = True))
        
        if writeFile:
            this.filesystem.add(filename, manifest, replace = True)
        
        return manifest



from ..gameobject import GameObject
from ..Utils.filesystem import *
from .widget import get_widget

class Layout(GameObject):
    def __init__(
        this,
        file  : str | bytes | File = None,
        filesystem: Filesystem | Folder = None,
        gamepath: str = None,
        assets: str = '/assets',
        baseassets: str = '/',
    ) -> None:
        super().__init__(filesystem, gamepath, assets, baseassets)
        
        
        this.filename = ''
        
        get_widget('test').widget
        
        



import io
import filetype
from PIL import Image

from .. import filesystem
from ..waltex import Waltex

class WaltexFile(filesystem.Reader):
    MIME = 'image/waltex'
    EXTENSION = 'waltex'
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        return mime == this.MIME
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        byte_order = 'little'
        if 'byte_order' in kwargs:
            byte_order = kwargs['byte_order']
        return Waltex(rawdata.getvalue(), byte_order=byte_order)
    
class ImageFile(filesystem.Reader):
    MIME = 'image/'
    EXTENSION = ''
    
    def check(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        newMime : str = filetype.guess_mime(rawdata.getvalue())
        
        return newMime != None and newMime.startswith(this.MIME)
    
    def read(this, mime: str, extension: str, rawdata: io.BytesIO, **kwargs):
        if WaltexFile().check(mime, extension, rawdata, **kwargs):
            return WaltexFile().read(mime, extension, rawdata, **kwargs)
        
        return Image.open(rawdata)

TYPES = [WaltexFile(), ImageFile()]



import lxml
from lxml import etree
from PIL import Image, ImageTk
from .widget import Widget, register_widget

class WT_SCROLLABLE_CAMERA(Widget):
    """
        Used for scrolling in `SN_MainMenu_v2.xml`
    """
    def __init__(this, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

register_widget('WT_SCROLLABLE_CAMERA', WT_SCROLLABLE_CAMERA,)



from lxml import etree
from PIL import Image, ImageTk
import numpy
from copy import deepcopy
import math

from .imagelist import Imagelist
from ..Utils.filesystem import *
from ..Utils.XMLTools import strbool
from ..gameobject import GameObject

class Sprite(GameObject):
    def __init__(
        this,
        file : str | bytes | File,
        filesystem: Filesystem | Folder = None,
        gamepath: str = None, assets: str = '/assets',
        baseassets : str = '/',
        properties : dict = {},
        scale : int = 10,
    ) -> None:
        """Game sprite.

        Args:
            this (_type_): _description_
            file (str | bytes | File): Sprite file.
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
            assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
            properties (dict, optional): Sprite properties. Defaults to {}.
            scale (int, optional): Sprite image scale. Defaults to 10.
        """
        
        super().__init__(filesystem, gamepath, assets, baseassets)
        this.file = super().get_file(file)
        
        this.xml : etree.ElementBase = etree.parse(this.file).getroot()
        
        this.properties = deepcopy(properties)
        this.animations : list[Sprite.Animation] = []
        
        this.scale = scale
        
        this.readXML()
        this.animation = 0
    
    def setAnimation(this, animation : str | int):
        """Set the current animation for the Sprite

        Args:
            animation (str | int): Animation name or index.
        """
        this.animation = animation
    
    @property
    def image(this) -> Image.Image:
        """Image of sprite

        Returns:
            PIL.Image.Image: PIL Image
        """
        image = this.animation.image.copy()
        gridSize = numpy.array(this.gridSize)
        size = gridSize * this.scale
        size = [abs(round(x)) for x in size]
        image = image.resize(size)
        
        image = image.rotate(this.angle, Image.BILINEAR, expand=True)
        
        return image
    
    @property
    def animation(this) -> 'Sprite.Animation':
        """Returns the current animation

        Returns:
            Sprite.Animation: A Sprite.Animation class
        """
        return this._currentAnimation
    @animation.setter
    def animation(this, animation : str | int):
        if isinstance(animation, (int, float)):
            animation = int(animation)
            this._currentAnimation = this.animations[animation]
        elif isinstance(animation, str):
            for a in this.animations:
                if a.name == animation:
                    this._currentAnimation = a
                    break
        elif isinstance(animation, this.Animation):
            this._currentAnimation = animation
    
    @property
    def frame(this) -> int:
        """The current animation frame.

        Returns:
            int: Current animation frame index.
        """
        return this.animation.frame
    @frame.setter
    def frame(this, value : int):
        this.animation.frame = value
    
    @property
    def filename(this) -> str:
        """Sprite filename
        """
        if 'filename' in this.properties:
            return this.properties['filename']
        else:
            return None
    @filename.setter
    def filename(this, value):
        this.properties['filename'] = value
        
    def readXML(this):
        """Read Sprite XML
        """
        this.animations = []
        for element in this.xml:
            if (not element is etree.Comment) or element.tag == 'Animation':
                animation = this.Animation(
                    element,
                    this.filesystem
                )
                
                this.animations.append(animation)
    
    def export(this, path : str = None):
        """Export the Sprite XML file

        Args:
            path (str, optional): Path to export into the filesystem. Defaults to the original filename.

        Raises:
            TypeError: Path is not a file.

        Returns:
            str: Contents of saved file.
        """
        xml : etree.ElementBase = etree.Element('Sprite')
        
        for animation in this.animations:
            xml.append(animation.getXML())
        
        this.xml = xml
        
        output = etree.tostring(xml, pretty_print=True, xml_declaration=True, encoding='utf-8')
        
        if path == None:
            if this.filename:
                path = this.filename
        
        if path != None:
            if (file := this.filesystem.get(path)) != None:
                if isinstance(file, File):
                    file.write(output)
                else:
                    raise TypeError(f'Path {path} is not a file.')
            else:
                this.filesystem.add(path, output)
        
        return output
    
    @property
    def visible(this) -> bool:
        """Whether this Sprite is visible or not
        """
        if 'visible' in this.properties:
            return strbool(this.properties['visible'])
        return False
    @visible.setter
    def visible(this, value : bool | str):
        this.properties['visible'] = str(strbool(value)).lower()
    
    @property
    def isBackground(this):
        """Whether this Sprite is a background
        """
        if 'isBackground' in this.properties:
            return strbool(this.properties['isBackground'])
        return False
    @isBackground.setter
    def isBackground(this, value : bool | str):
        this.properties['isBackground'] = str(strbool(value)).lower()
    
    @property
    def gridSize(this) -> tuple[float,float]:
        """The gridSize (size) of this Sprite

        Returns:
            tuple[float,float]: (width,height)
        """
        if 'gridSize' in this.properties:
            return tuple([float(x) for x in this.properties['gridSize'].split()])
        return (1,1)
    @gridSize.setter
    def gridSize(this, value : tuple[int,int] | str):
        if isinstance(value, str):
            this.properties['gridSize'] = value
        elif isinstance(value, (tuple, list)):
            this.properties['gridSize'] = ' '.join([str(x) for x in value])
    
    @property
    def pos(this) -> tuple[float,float]:
        """Position of Sprite relative to the center of the Object

        Returns:
            tuple[float,float]: (x,y)
        """
        if 'pos' in this.properties:
            return tuple([float(x) for x in this.properties['pos'].split()])
        return (0,0)
    @pos.setter
    def pos(this, value : tuple[int,int] | str):
        if isinstance(value, str):
            this.properties['pos'] = value
        elif isinstance(value, (tuple, list)):
            this.properties['pos'] = ' '.join([str(x) for x in value])
    
    @property
    def angle(this) -> float:
        """Sprite rotation angle

        Returns:
            float: Angle as degrees
        """
        if 'angle' in this.properties:
            return float(this.properties['angle'])
        return 0
    @angle.setter
    def angle(this, value : int | float):
        this.properties['angle'] = str(value)

    class Animation(GameObject):
        def __init__(
            this,
            xml : str | etree.ElementBase,
            filesystem: Filesystem | Folder = None,
            gamepath: str = None,
            assets: str = '/assets',
            baseassets: str = '/',
        ) -> None:
            """Animation for Sprite.

            Args:
                xml (str | etree.Element): lxml.etree Element xml element for sprite.
                filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
                gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
                assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
                baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
            """
            super().__init__(filesystem, gamepath, assets, baseassets)
            
            if isinstance(xml, str):
                this.xml : etree.ElementBase = etree.parse(xml).getroot()
            else:
                this.xml = xml
            
            this.properties = {}
            this.name = ''
            this.textureBasePath = '/Textures/'
            this.atlas : Imagelist = None
            this.fps = 30
            this.playbackMode = 'ONCE'
            this.loopCount = 1
            
            this._PhotoImage = None
            
            this.frame = 0
            
            this.frames : list[Sprite.Animation.Frame] = []
            
            this.readXML()
        
        @property
        def image(this) -> Image.Image:
            """Current Animation image

            Returns:
                PIL.Image.Image: PIL Image
            """
            if this.frame > len(this.frames):
                this.frame = 0
            if this.frame < 0:
                this.frame = len(this.frames)
            
            return this.frames[this.frame].image
        
        @property
        def PhotoImage(this) -> ImageTk.PhotoImage:
            """Tkinter PhotoImage for the Animation
            """
            this._PhotoImage = ImageTk.PhotoImage(this.image)
            return this._PhotoImage
            
        def readXML(this):
            """Read the xml for this Animation
            """
            this.getAttributes()
            this.getFrames()
        
        def getAttributes(this):
            """Get all the attributes of this Animation
            """
            this.properties = this.xml.attrib
            
            if 'name' in this.properties:
                this.name = this.properties['name']
            if 'textureBasePath' in this.properties:
                this.textureBasePath = this.properties['textureBasePath']
            if 'atlas' in this.properties:
                this.atlasPath = this.properties['atlas']
                this.atlas = Imagelist(
                    this.filesystem.get(this.properties['atlas']),
                    this.filesystem,
                    HD=False
                )
                
                # this.atlasHD = Imagelist(
                #     this.filesystem.get(this.properties['atlas']),
                #     this.filesystem,
                #     HD=True
                # )
                
            if 'fps' in this.properties:
                this.fps = float(this.properties['fps'])
            if 'playbackMode' in this.properties:
                this.playbackMode = this.properties['playbackMode']
            if 'loopCount' in this.properties:
                this.loopCount = int(this.properties['loopCount'])
            
            
        def getFrames(this) -> list['Sprite.Animation.Frame']:
            """Get a list of all the Animation `Frame`s

            Returns:
                list[Sprite.Animation.Frame]: List of all the Frames in this Animation.
            """
            this.frames = []
            
            if this.xml == None:
                return None
            for f in this.xml:
                if (not f is etree.Comment) and f.tag == 'Frame':
                    this.frames.append(this.Frame(
                        f.attrib,
                        this.atlas,
                        this.textureBasePath
                    ))
            
            return this.frames

        def updateProperties(this):
            """Update the Sprite properties
            """
            def updateProperty(property : str, value, default = None):
                if default != None and value == default:
                    if property in this.properties:
                        del this.properties[property]
                else:
                    this.properties[property] = value
            
            updateProperty('name', this.name)
            updateProperty('textureBasePath', this.textureBasePath)
            this.atlas.export(this.atlasPath, exportImage=True)
            updateProperty('atlas', this.atlasPath)
            updateProperty('fps', str(this.fps))
            updateProperty('playbackMode', this.playbackMode)
            updateProperty('loopCount', str(this.loopCount))
        
        def getXML(this):
            """Get the XML of this Animation

            Returns:
                etree.Element: etree Element
            """
            this.updateProperties()
            xml : etree.ElementBase = etree.Element('Animation', **this.properties)
            
            for frame in this.frames:
                xml.append(frame.getXML())
            
            this.xml = xml
            return this.xml
        
        
        # Frame
        class Frame():
            _offset = (0,0)
            _scale = (1,1)
            _angleDeg = 0
            _repeat = 1
            def __init__(
                this,
                properties : dict,
                atlas : Imagelist = None,
                textureBasePath : str = None,
            ) -> None:
                """Frame for Sprite.Animation.

                Args:
                    this (_type_): _description_
                    properties (dict): Image properties.
                    atlas (Imagelist, optional): Image atlas for Image. Defaults to None.
                    textureBasePath (str, optional): Directory to put image in. Defaults to None.
                """
                this.atlas = atlas
                this.textueBasePath = textureBasePath
                this.properties = properties
                
                this.name = ''
                this.offset = this._offset
                this.scale = this._scale
                this.angleDeg = this._angleDeg
                this.repeat = this._repeat
                
                
                
                this.getData()
                this.getImage()
            
            def getData(this):
                """Get the Frame data
                """
                if 'name' in this.properties:
                    this.name = this.properties['name']
                if 'offset' in this.properties:
                    this.offset = tuple([float(x) for x in this.properties['offset'].split()])
                if 'scale' in this.properties:
                    this.scale = tuple([float(x) for x in this.properties['scale'].split()])
                if 'angleDeg' in this.properties:
                    this.angleDeg = float(this.properties['angleDeg'])
                if 'repeat' in this.properties:
                    this.repeat = int(this.properties['repeat'])
                
            def getImage(this):
                this._image = this.atlas.getImage(this.name)
            
            @property
            def image(this) -> Image.Image:
                """Image of this Image

                Returns:
                    PIL.Image.Image: PIL Image
                """
                if this._image:
                    image = this._image.image.copy()
                    image = image.resize(tuple([round(_) for _ in (numpy.array(this._image.size) * numpy.array(this.scale))]))
                    image = image.rotate(this.angleDeg, expand = True)
                else:
                    image = Image.new('RGBA', (1,1), (0,0,0,0))
                return image
            @image.setter
            def image(this, image : Image.Image):
                this._image = image
            
            def updateProperties(this):
                """Update Image properties
                """
                def updateProperty(property : str, value, default):
                    if value == default:
                        if property in this.properties:
                            del this.properties[property]
                    else:
                        this.properties[property] = value
                
                this.properties['name'] = this.name
                
                updateProperty(
                    'offset',
                    ' '.join([str(_) for _ in this.offset]),
                    ' '.join([str(_) for _ in this._offset])
                )
                
                updateProperty(
                    'scale',
                    ' '.join([str(_) for _ in this.scale]),
                    ' '.join([str(_) for _ in this._scale])
                )
                
                updateProperty(
                    'angleDeg',
                    str(this.angleDeg),
                    str(this._angleDeg)
                )
                
                updateProperty(
                    'repeat',
                    str(this.repeat),
                    str(this._repeat)
                )

            def getXML(this) -> etree.ElementBase:
                """Get the XML for the Frame

                Returns:
                    etree.Element: XML of this Frame
                """
                this.updateProperties()
                return etree.Element('Frame', **this.properties)



import os
import pathlib

def joinPath(*args : str) -> str:
    """Join a list of paths to create a new path, while ignoring any forward or backward slashes (`/` `\`)

    Args:
        *args: (str): str path
    Returns:
        str: New path
    """
    
    if len(args) == 0:
        return ''

    path = pathlib.Path(args[0], *[makeRelativePath(p) for p in args[1:]])
    return path.as_posix()

def makeRelativePath(path : str) -> str:
    """Remove forward and backward slashes (`/` `\`) from the beginning of a path.

    Args:
        path (str): Path to check

    Returns:
        str: New path without a forward or backward slash.
    """
    if path == '':
        return '.'
    parts = pathlib.Path(path).parts
    if len(parts) == 0:
        return path
    
    if parts[0] in ['\\', '/']:
        return pathlib.Path(*parts[1::]).as_posix()
    return path


# main widget class
from ...Utils.filesystem import Filesystem, Folder
from ...gameobject import GameObject


from PIL import Image
from lxml import etree

WIDGETS : dict[str, 'Widget'] = {}

class Widget(GameObject):
    def __init__(
        this,
        xml : etree.ElementBase = None,
        filesystem: Filesystem | Folder = None,
        gamepath: str = None,
        assets: str = '/assets',
        baseassets: str = '/',
        screenSize : tuple = (900,720),
    ) -> None:
        """
            Main widget
        """
        if xml == None:
            return
        
        
        super().__init__(filesystem, gamepath, assets, baseassets)
        
        this.xml = xml
        this.properties = this.xml.attrib
        this.type = this.properties['type']

        this.pos = (0, 0)
        this.size = (0, 0)
        this.id = 0
        this.layer = 0

        this.forceAspect = False
        this.visible = True

        this.getValues()

        this.image = Image.new('RGBA', (100,100))

    def getValues(this):
        if 'pos' in this.properties:
            this.pos = [float(v) for v in tuple(this.properties['pos'].split(' '))]

        if 'id' in this.properties:
            this.id = float(this.properties['id'])

        if 'layer' in this.properties:
            this.layer = float(this.properties['layer'])

        if 'size' in this.properties:
            this.size = [float(v) for v in tuple(this.properties['size'].split(' '))]

        if 'forceAspect' in this.properties:
            this.setForceAspect(this.properties['forceAspect'])

        if 'visible' in this.properties:
            this.visible = bool(this.properties['visible'])

    def setForceAspect(this, aspect = (1,1)):
        if isinstance(aspect, str):
            forceAspect = tuple([float(v) for v in aspect.split(':')])
        elif not aspect:
            forceAspect = False
        else:
            forceAspect = tuple(aspect)

        this.forceAspect = forceAspect

    @property
    def type(this):
        return this.properties['type']
    @type.setter
    def type(this, value):
        this.properties['type'] = value

        if this.properties['type'] in WIDGETS:
            this.__class__ = WIDGETS[this.properties['type']]
        else:
            this.__class__ = Widget


def register_widget(name : str, class_ : Widget):

    if not isinstance(class_, type):
        class_ = class_.__class__

    if not isinstance(class_(None), Widget):
        raise TypeError('class has to be inherited by Widget')
    if not isinstance(name, str):
        raise TypeError('name must be a string')
    try:
        if WIDGETS[name]:
            raise NameError(f'widget "{name}" already exists')
    except:
        pass
    WIDGETS[name] = class_

def get_widget(name, *args, **kwargs) -> Widget:
    if name in WIDGETS:
        return WIDGETS[name](**args, **kwargs)
    else:
        return Widget(**args, **kwargs)



from .Utils.filesystem import *

import io
import typing
import numpy
import math

class GameObject():
    def __init__(
        this,
        filesystem : Filesystem | Folder = None,
        gamepath : str = None,
        assets : str = '/assets',
        baseassets : str = '/',
    ) -> None:
        """Load filesystem

        Args:
            filesystem (Filesystem | Folder, optional): Filesystem to use. Defaults to None.
            gamepath (str, optional): Game path. Only used if filesystem not specified. Defaults to None.
            assets (str, optional): Assets path relative to game path. Only used if filesystem not specified. Defaults to '/assets'.
            baseassets (str, optional): Base assets path within the assets folder, e.g. `/perry/` in wmp. Defaults to `/`
        """
        this.gamepath = gamepath
        this.assets = assets
        if this.assets == None:
            this.assets = '/assets'
        this.baseassets = baseassets
        
        # try:
        this.filesystem = filesystem
        if isinstance(this.filesystem, Filesystem):
            this.gamepath = this.filesystem.gamepath
            this.assets = this.filesystem.assets
            
        elif isinstance(this.filesystem, Folder):
            pass
        
        elif hasattr(this.filesystem, 'filesystem'):
            this.filesystem = this.filesystem.filesystem

        else:
            this.filesystem = Filesystem(this.gamepath, this.assets)
            this.filesystem.getAssets()
        # except Exception as e:
        #     print(f'Error: {str(e)}')
        #     raise FileNotFoundError('Must have a valid `filesystem` or `gamepath`')
    
    def get_file(
        this,
        file : bytes | File | io.BytesIO | str,
        template : str | io.BytesIO = None,
    ) -> io.BytesIO | str:
        """Get file

        Args:
            file (bytes | wmwpy.Utils.filesystem.File | io.BytesIO | str): Content of file. Can be bytes, wmwpy File, str (contents of file) or file-like object.
            template (str | io.BytesIO, optional): Fallback for file if file == None. Defaults to ''

        Raises:
            TypeError: File can only be 'str', 'bytes', or file-like object.

        Returns:
            io.BytesIO | str: New file object or str.
        """
        if isinstance(file, bytes):
            fileio = io.BytesIO(file)
        elif isinstance(file, File):
            fileio = file.rawdata
        elif file == None:
            if template != None:
                return this.get_file(template)
            else:
                return None
        elif not hasattr(file, 'read') and not isinstance(file, str):
            raise TypeError(f"file can only be 'str', 'bytes', or file-like object.")
        else:
            return file
    
        return fileio
    
    def truePos(
        this,
        pos : tuple[int,int] = (0,0),
        obj_size : tuple[int,int] = (0,0),
        parent_size : tuple[int,int] = (0,0),
        offset : tuple[int,int] = (0,0),
        obj_anchor : typing.Literal['center', 'c', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'center',
        parent_anchor : typing.Literal['center', 'c', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'center',
        scale : int = 1,
    ) -> tuple[int,int]:
        """Get the true position of an object inside a larger object.

        Args:
            pos (tuple[x,y], optional): The position of the object. Defaults to (0,0).
            obj_size (tuple[width,height], optional): The size of the object. Defaults to (0,0).
            parent_size (tuple[width,height], optional): The size of the parent. Defaults to (0,0).
            offset (tuple[x,y], optional): An offset. Defaults to (0,0).
            obj_anchor (typing.Literal["center", "c", "n", "ne", "e", "se", "s", "sw", "w", "nw"], optional): The object anchor. Defaults to 'center'.
            parent_anchor (typing.Literal["center", "c", "n", "ne", "e", "se", "s", "sw", "w", "nw"], optional): The parent anchor. Defaults to 'center'.
            scale (int, optional): A scale for the position. Defaults to 1.

        Raises:
            TypeError: Invalid anchor.

        Returns:
            tuple[int,int]: The new position.
        """
        pos = numpy.array(pos)
        obj_size = numpy.array(obj_size)
        parent_size = numpy.array(parent_size)
        offset = numpy.array(offset)
        
        anchors = {
            'center': numpy.array((0.5,0.5)),
            'c': numpy.array((0.5,0.5)),
            'nw': numpy.array((0,0)),
            'n': numpy.array((0.5,0)),
            'ne': numpy.array((1,0)),
            'e': numpy.array((1,0.5)),
            'se': numpy.array((1,1)),
            's': numpy.array((0.5,1)),
            'sw': numpy.array((0,1)),
            'w': numpy.array((0,0.5)),
        }
        
        obj_anchor = obj_anchor.lower()
        parent_anchor = parent_anchor.lower()
        
        if not obj_anchor in anchors:
            raise TypeError(f"Anchor '{obj_anchor}' not supported")
        if not parent_anchor in anchors:
            raise TypeError(f"Anchor '{parent_anchor}' not supported")
        
        obj_anchor = anchors[obj_anchor]
        parent_anchor = anchors[parent_anchor]
        
        pos = pos * [1,-1]
        obj_size = obj_size * [1,-1]
        offset = offset * [1,-1]
        pos = ((pos - offset) - ((obj_size * obj_anchor) - (parent_size * parent_anchor))) * scale
        
        return tuple(pos)



import lxml
from lxml import etree
from PIL import Image, ImageTk
from ..Utils.waltex import WaltexImage
from .widget import Widget

class Widgets():
    def __init__(this, element : etree.Element, gamePath : str, screenSize = (), texturePath : str = None, baseLayoutFile : str = None) -> None:
        this.element = element
        
        this.attributes = this.element.attrib
        
        if not texturePath:
            texturePath = this.attributes['texturePath']
        this.texturePath = texturePath
        
        if not baseLayoutFile:
            baseLayoutFile = this.attributes['baseLayoutFile']
        this.baseLayoutFile = baseLayoutFile
        
        this.gamePath = gamePath
        
        this.widgets = []
        this.comments = []
        
        this.getWidgets()
        
    def getWidgets(this):
        for w in this.element:
            if not isinstance(w, etree.Comment):
                widget = Widget(w, this.texturePath)
                this.widgets.append(widget)
            else:
                this.comments.append(w)
        
    


import os
from lxml import etree
from .waltex import WaltexImage, Waltex
from PIL import Image

from .path import joinPath
from .XMLTools import findTag
from .filesystem import Filesystem, Folder, File

_cachedWaltextImages = {}

class Texture():
    def __init__(this, image : Image.Image | Waltex | File) -> None:
        """Texture for image.

        Args:
            image (Image.Image | Waltex | File): Image object. Can be PIL.Image.Image, Waltex image, or file.

        Raises:
            TypeError: image must be PIL.Image.Image, Waltex, or filesystem.File.
        """
        this._image = image
        
        if isinstance(this._image, File):
            this._image = this._image.read()
        
        if isinstance(this._image, Waltex):
            this.image = this._image.image
        elif isinstance(this._image, Image.Image):
            this.image = this._image
        else:
            raise TypeError('image must be PIL.Image.Image, Waltex, or filesystem.File.')
        
    @property
    def size(this):
        return this.image.size

def getHDFile(file : str) -> str:
    """Get HD filename.

    Args:
        file (str): Filename

    Returns:
        str: HD filename.
    """
    split = os.path.splitext(file)
    split = list(split)
    split.insert(1, '-HD')
    return ''.join(split)
            
def getTexture(path : str, textureSettings : dict, size : tuple, cache = True) -> Image.Image:
    """Get image.

    Args:
        path (str): Path to Image.
        textureSettings (dict): Texture settings.
        size (tuple[width,height]): Size of image.
        cache (bool, optional): Whether to cache waltex images. Defaults to True.

    Returns:
        Image.Image: PIL Image.
    """
    type = os.path.splitext(path)[1][1:]
    image = None
    if type == 'waltex':
        if cache:
            try:
                image = _cachedWaltextImages[path].copy()
            except:
                pass
        if image == None:
            image = Waltex(
                path
            ).image
            if cache:
                _cachedWaltextImages[path] = image.copy()
    else:
        image = Image.open(path).convert('RGBA')
        
    return image
    
def getTextueSettings(gamepath : str, assets : str, textureSettings : str, name : str, ) -> dict:
    """Get the texture settings of a texture.
    
    Really needs an update, as now I can reliably get the path to the file using the `baseassets` property in the `Game` object.

    Args:
        gamepath (str): Path to game
        assets (str): Relative path from the gamepath to the assets folder.
        textureSettings (str): Path to `TextureSettings.xml` file
        name (str): Name of image to look for.

    Returns:
        dict: The texture settings as a dict.
    """
    fullpath = joinPath(gamepath, assets, textureSettings)
    xml = etree.parse(fullpath).getroot()
    
    Texture = etree.ElementBase()
    for i in xml:
        if not i.tag is etree.Comment:
            if i.tag == 'Texture' and i.get('name') == name:
                Texture = i
                break
            
    attributes = {
        'colorspace': 'RGBA4444',
        'premultiplyAlpha': False,
        'dePremultiplyAlpha': False,
    }
    
    values = Texture.attrib
    
    for key in values:
        attributes[key] = values[key]
    
    return attributes



from .textures import Texture
from . import textures
from .waltex import WaltexImage, Waltex
from . import XMLTools
from .path import joinPath
from . import filesystem
from .filesystem import Filesystem
from . import Types
